"""Functions for calculating for dry matter intake (DMI).

Includes equations for animals of different physiological states. Some equations
will use diet information in the prediction.
"""

import math
from typing import Union

import pandas as pd

import nasem_dairy.nasem_equations.nutrient_intakes as diet


# Precalculation for heifer DMI predicitons
def calculate_Kb_LateGest_DMIn(Dt_NDF: float) -> float:
    """
    Calculate the ______________ for predicting dry matter intake (DMI) in late gestation for heifers.

    Parameters
    ----------
    Dt_NDF : float
        The neutral detergent fiber (NDF) as a percentage.

    Returns
    -------
    float
        The precalculation factor (Kb_LateGest_DMIn) for heifer DMI predictions.

    Notes
    -----
    - This function is used as a _____ for predicting heifer dry matter intake (DMI).
    - The input Dt_NDF is constrained to the range of 30 to 55% of dry matter (DM).
    - Kb_LateGest_DMIn is calculated as -(0.365 - 0.0028 * Dt_NDF_drylim), where Dt_NDF_drylim is the constrained value.

    Examples
    --------
    Calculate the precalculation factor for heifer DMI predictions in late gestation:

    ```{python}
    import nasem_dairy as nd
    nd.calculate_Kb_LateGest_DMIn(Dt_NDF=40)
    ```
    """
    Dt_NDF_drylim = Dt_NDF  # Dt_NDF_drylim only used in this calculation
    if Dt_NDF < 30:  # constrain Dt_NDF to the range of 30 to 55% of DM
        Dt_NDF_drylim = 30
    if Dt_NDF > 55:
        Dt_NDF_drylim = 55

    Kb_LateGest_DMIn = -(0.365 - 0.0028 * Dt_NDF_drylim)
    return Kb_LateGest_DMIn


# Precalculation for heifer DMI predicitons
def calculate_An_PrePartWklim(An_PrePartWk: float) -> Union[int, float]:
    # Late Gestation eqn. from Hayirli et al., 2003) for dry cows and heifers
    if An_PrePartWk < -3:  # constrain to the interval 0 to -3.
        An_PrePartWklim = -3
    elif An_PrePartWk > 0:
        An_PrePartWklim = 0
    else:
        An_PrePartWklim = An_PrePartWk
    return An_PrePartWklim


# Need when DMIn_eqn == 2,3,4,5,6,7,10
def calculate_Dt_DMIn_BW_LateGest_i(
    An_PrePartWklim: Union[int, float], 
    Kb_LateGest_DMIn: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Ka_LateGest_DMIn": 1.47, "Kc_LateGest_DMIn": -0.035}
    
    calculate_Dt_DMIn_BW_LateGest_i(
        An_PrePartWklim = 6.0, Kb_LateGest_DMIn = 0.05, coeff_dict = coeff_dict
    )
    ```
    """
    # Late gestation individual animal prediction, % of BW.  Use to assess for a
    # specific day for a given animal
    Dt_DMIn_BW_LateGest_i = (coeff_dict['Ka_LateGest_DMIn'] + 
                             Kb_LateGest_DMIn * An_PrePartWklim + 
                             coeff_dict['Kc_LateGest_DMIn'] * An_PrePartWklim**2)
    return Dt_DMIn_BW_LateGest_i


# Need when DMIn_eqn == 10,12,13,14,15,16,17
def calculate_Dt_DMIn_BW_LateGest_p(
    An_PrePartWkDurat: Union[int, float], 
    Kb_LateGest_DMIn: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {
        "Ka_LateGest_DMIn": 1.47, "Kc_LateGest_DMIn": -0.035
    }
    
    calculate_Dt_DMIn_BW_LateGest_p(
        An_PrePartWkDurat = 6.0, Kb_LateGest_DMIn = 0.05, 
        coeff_dict = coeff_dict
    )
    ```
    """
    # Late gestation Group/Pen mean DMI/BW for an interval of 0 to 
    # PrePart_WkDurat. Assumes pen steady state and PrePart_wk = pen mean
    Dt_DMIn_BW_LateGest_p = (coeff_dict['Ka_LateGest_DMIn'] * An_PrePartWkDurat
                             + Kb_LateGest_DMIn / 2 * An_PrePartWkDurat**2 +
                             coeff_dict['Kc_LateGest_DMIn'] / 3 *
                             An_PrePartWkDurat**3) / An_PrePartWkDurat
    return Dt_DMIn_BW_LateGest_p


# Need when DMIn_eqn == 2,3,4,5,6,7
def calculate_Dt_DMIn_Heif_LateGestInd(
    An_BW: float, 
    Dt_DMIn_BW_LateGest_i: float
) -> float:
    # Individual intake for the specified day prepart or the pen mean intake 
    # for the interval, 0 to PrePart_WkDurat
    Dt_DMIn_Heif_LateGestInd = 0.88 * An_BW * Dt_DMIn_BW_LateGest_i / 100 
    # Individual animal
    return Dt_DMIn_Heif_LateGestInd


# Need when DMIn_eqn == 12,13,14,15,16,17
def calculate_Dt_DMIn_Heif_LateGestPen(
    An_BW: float, 
    Dt_DMIn_BW_LateGest_p: float
) -> float:
    Dt_DMIn_Heif_LateGestPen = 0.88 * An_BW * Dt_DMIn_BW_LateGest_p / 100  
    # Pen mean
    return Dt_DMIn_Heif_LateGestPen


# Need when DMIn_eqn == 5,7,15,17
def calculate_Dt_NDFdev_DMI(An_BW: float, Dt_NDF: float) -> float:
    # NRC 2020 Heifer Eqns. from the Transition Ch., Line 316
    Dt_NDFdev_DMI = Dt_NDF - (23.11 + 0.07968 * An_BW - 0.00006252 * An_BW**2)
    return Dt_NDFdev_DMI


# DMIn_eqn == 2, 12
def calculate_Dt_DMIn_Heif_NRCa(An_BW: float, An_BW_mature: float) -> float:
    '''
    Test docs for calcualte_Dt_DMIn_Heif_NRCa
    '''
    # Animal factors only, eqn. 2-3 NRC
    Dt_DMIn_Heif_NRCa = (0.022 * An_BW_mature * 
                         (1 - math.exp(-1.54 * An_BW / An_BW_mature)))
    return Dt_DMIn_Heif_NRCa


# DMIn_eqn == 3, 13
def calculate_Dt_DMIn_Heif_NRCad(
    An_BW: float, 
    An_BW_mature: float, 
    Dt_NDF: float
) -> float:
    # Anim & diet factors, eqn 2-4 NRC
    Dt_DMIn_Heif_NRCad = ((0.0226 * An_BW_mature * 
                           (1 - math.exp(-1.47 * An_BW / An_BW_mature))) - 
                           (0.082 * (Dt_NDF - (23.1 + 56 * An_BW / An_BW_mature) 
                                     - 30.6 * (An_BW / An_BW_mature)**2)))
    return Dt_DMIn_Heif_NRCad


# DMIn_eqn == 4, 14
def calculate_Dt_DMIn_Heif_H1(An_BW: float) -> float:
    # Holstein, animal factors only
    Dt_DMIn_Heif_H1 = 15.36 * (1 - math.exp(-0.0022 * An_BW))
    return Dt_DMIn_Heif_H1


# DMIn_eqn == 5, 15
def calculate_Dt_DMIn_Heif_H2(An_BW: float, Dt_NDFdev_DMI: float) -> float:
    # Holstein, animal factors and NDF
    Dt_DMIn_Heif_H2 = (15.79 * (1 - math.exp(-0.0021 * An_BW)) - 
                       (0.082 * Dt_NDFdev_DMI))
    return Dt_DMIn_Heif_H2


# DMIn_eqn == 6, 16
def calculate_Dt_DMIn_Heif_HJ1(An_BW: float) -> float:
    """
    _summary_

    Parameters
    ----------
    An_BW : _type_
        _description_

    Returns
    -------
    _type_
        _description_
    """
    #Holstein x Jersey, animal factors only
    Dt_DMIn_Heif_HJ1 = 12.91 * (1 - math.exp(-0.00295 * An_BW))
    return Dt_DMIn_Heif_HJ1


# DMIn_eqn == 7, 17
def calculate_Dt_DMIn_Heif_HJ2(An_BW:float, Dt_NDFdev_DMI: float) -> float:
    """
    Calculate the predicted dry matter intake (DMI) for Holstein x Jersey crossbred heifers
    considering animal factors and neutral detergent fiber (NDF).

    Parameters
    ----------
    An_BW : float
        The body weight of the heifer in kg.
    Dt_NDFdev_DMI : float
        The neutral detergent fiber (NDF) as a percentage

    Returns
    -------
    float
        The predicted dry matter intake (DMI) in kg

    Notes
    -----
    - See equation number ___ in the Nutrient Requirements of Dairy Cattle book (NASEM, 2021)
    - See line number 317 in the original R code published with the book's software
    - This function is equated when equation_selection for DMIn_eqn is equal to 7 and 17


    Examples
    -------

    ```{python}
    import nasem_dairy as nd
    nd.calculate_Dt_DMIn_Heif_HJ2(
        An_BW = 700,
        Dt_NDFdev_DMI = 14
        )
    ```
   
    """
    #Holstein x Jersey, animal factors and NDF
    Dt_DMIn_Heif_HJ2 = (13.48 * (1 - math.exp(-0.0027 * An_BW)) - 
                        (0.082 * Dt_NDFdev_DMI))
    return Dt_DMIn_Heif_HJ2


# DMIn_eqn == 8
def calculate_Dt_DMIn_Lact1(
    An_BW: float, 
    An_BCS: float, 
    An_LactDay: int,
    An_Parity_rl: int, 
    Trg_NEmilkOut: float
) -> float:
    """
    Calculate the predicted dry matter intake (DMI) for lactating dairy cows using equation 8.

    Parameters
    ----------
    An_BW : float
        The body weight of the lactating cow in kg.
    An_BCS : float
        The body condition score (BCS) of the lactating cow.
    An_LactDay : int
        The lactation day of the cow.
    An_Parity_rl : int
        The parity of the cow (number of times calved).
    Trg_NEmilkOut : float
        Net energy of milk production in Mcal.

    Returns
    -------
    float
        The predicted dry matter intake (DMI) in kg.

    Notes
    -----
    - See equation number 2-1 in the Nutrient Requirements of Dairy Cattle book (NASEM, 2021).
    - See lines 387-92 in the original R code published with the book's software.
    - Trg_NEmilkOut is calculated as Trg_NEmilk_Milk * Trg_MilkProd (see line 386).
    - This function is associated with DMIn_eqn equal to 8.

    Examples
    --------
    Calculate the dry matter intake for a lactating dairy cow:

    ```python
    import nasem_dairy as nd
    nd.calculate_Dt_DMIn_Lact1(
        An_BW=600,
        An_BCS=3.5,
        An_LactDay=120,
        An_Parity_rl=2,
        Trg_NEmilkOut=0.65
    )
    ```

    Returns
    -------
    float
        The predicted dry matter intake (DMI) for the given parameters.
    """
    Dt_DMIn_Lact1 = ((3.7 + 5.7 * (An_Parity_rl - 1) + 0.305 * Trg_NEmilkOut + 
                      0.022 * An_BW + 
                      (-0.689 - 1.87 * (An_Parity_rl - 1)) * An_BCS) * 
                      (1 - (0.212 + 0.136 * (An_Parity_rl - 1)) * 
                       math.exp(-0.053 * An_LactDay))) # Line 390
    return Dt_DMIn_Lact1

# DMIn_eqn == 9 
def calculate_Dt_DMIn_Lact2(
    Dt_ForNDF: float, 
    Dt_ADF: float, 
    Dt_NDF: float, 
    Dt_ForDNDF48_ForNDF: float,
    Trg_MilkProd: float
) -> float:
    Dt_DMIn_Lact2 = (12.0 - 0.107 * Dt_ForNDF + 8.17 * Dt_ADF / Dt_NDF + 
                     0.0253 * Dt_ForDNDF48_ForNDF - 
                     0.328 * (Dt_ADF / Dt_NDF - 0.602) * 
                     (Dt_ForDNDF48_ForNDF - 48.3) + 
                     0.225 * Trg_MilkProd + 0.00390 * 
                     (Dt_ForDNDF48_ForNDF - 48.3) * (Trg_MilkProd - 33.1))
    return Dt_DMIn_Lact2

# DMIn_eqn == 10
def calculate_Dt_DMIn_DryCow1_FarOff(
    An_BW: float, 
    Dt_DMIn_BW_LateGest_i: float
) -> float:
    Dt_DMIn_DryCow1_FarOff = An_BW * Dt_DMIn_BW_LateGest_i / 100
    return Dt_DMIn_DryCow1_FarOff


# DMIn_eqn == 10
def calculate_Dt_DMIn_DryCow1_Close(
    An_BW: float, 
    Dt_DMIn_BW_LateGest_p: float
) -> float:
    Dt_DMIn_DryCow1_Close = An_BW * Dt_DMIn_BW_LateGest_p / 100
    return Dt_DMIn_DryCow1_Close


def calculate_Dt_DMIn_DryCow_AdjGest(
    An_GestDay: int,
    An_GestLength: int,
    An_BW: float
) -> float:
    # from Hayirli et al., 2003 JDS
    if (An_GestDay - An_GestLength) < -21:
        Dt_DMIn_DryCow_AdjGest = 0
    else:
        Dt_DMIn_DryCow_AdjGest = (
            An_BW * (-0.756 * math.exp(0.154 * (An_GestDay - An_GestLength))) 
            / 100
            )
    return Dt_DMIn_DryCow_AdjGest


# DMIn_eqn == 11
def calculate_Dt_DMIn_DryCow2(
    An_BW: float, 
    Dt_DMIn_DryCow_AdjGest: float
) -> float:
    Dt_DMIn_DryCow2 = An_BW * 1.979 / 100 + Dt_DMIn_DryCow_AdjGest
    return Dt_DMIn_DryCow2


def calculate_Dt_DMIn_Calf1(
    Dt_DMIn_ClfLiq: float, 
    Dt_DMIn_ClfStrt: float,
    Dt_DMIn_ClfFor: float
) -> float:
    """
    Dt_DMIn_Calf1: Calf DMI with predicted starter intake, kg/d
    Liquid feed + calfd starter + forage intake
    """
    Dt_DMIn_Calf1 = Dt_DMIn_ClfLiq + Dt_DMIn_ClfStrt + Dt_DMIn_ClfFor
    # DMI w/ predicted starter, Line 308
    return Dt_DMIn_Calf1


def calculate_Dt_DMIn(
    DMIn_eqn: int,
    Trg_Dt_DMIn: float,
    An_StatePhys: str,
    An_BW: float,
    An_BW_mature: float,
    An_BCS: int,
    An_LactDay: int,
    An_Parity_rl: int, 
    Trg_MilkProd: float,
    An_GestDay: int,
    An_GestLength: int,
    An_AgeDryFdStart: int,
    Env_TempCurr: float,
    An_PrePartWk: float,
    Trg_NEmilkOut: float,
    An_PrePartWklim: float,
    An_PrePartWkDurat: float,
    Fd_NDF: pd.Series,
    Fd_DMInp: pd.Series,
    Fd_ADF: pd.Series,
    Fd_ForNDF: pd.Series,
    Fd_Conc: pd.Series,
    Fd_DNDF48_input: pd.Series,
    Trg_Fd_DMIn: pd.Series,
    Fd_Category: pd.Series,
    Fd_CP: pd.Series,
    Fd_FA: pd.Series,
    Fd_Ash: pd.Series,
    Fd_St: pd.Series,
    coeff_dict: dict
) -> float:   
    Dt_NDF = (Fd_NDF * Fd_DMInp).sum()
    Dt_ADF = (Fd_ADF * Fd_DMInp).sum()
    Dt_ForNDF = (Fd_ForNDF * Fd_DMInp).sum()
    Fd_DNDF48 = diet.calculate_Fd_DNDF48(
        Fd_Conc, Fd_DNDF48_input
        )
    Dt_ForDNDF48 = diet.calculate_Dt_ForDNDF48(
        Fd_DMInp, Fd_Conc, Fd_NDF, 
        Fd_DNDF48
        )
    Dt_ForDNDF48_ForNDF = diet.calculate_Dt_ForDNDF48_ForNDF(
        Dt_ForDNDF48, Dt_ForNDF
        )
    Kb_LateGest_DMIn = calculate_Kb_LateGest_DMIn(Dt_NDF)

    if DMIn_eqn == 0:
        Dt_DMIn = Trg_Dt_DMIn

    elif DMIn_eqn == 1:
        Fd_DMIn_ClfLiq = diet.calculate_Fd_DMIn_ClfLiq(
            An_StatePhys, Trg_Fd_DMIn, Fd_Category
        )
        Fd_DMIn_ClfFor = diet.calculate_Fd_DMIn_ClfFor(
            Trg_Dt_DMIn, Fd_Conc, Fd_DMInp
        )
        Fd_GE = diet.calculate_Fd_GE(
            An_StatePhys, Fd_Category, Fd_CP, 
            Fd_FA, Fd_Ash, Fd_St,
            Fd_NDF, coeff_dict
        )
        Fd_DE_ClfLiq = diet.calculate_Fd_DE_ClfLiq(
            An_StatePhys, Fd_Category, Fd_GE
        )
        Fd_ME_ClfLiq = diet.calculate_Fd_ME_ClfLiq(
            An_StatePhys, Fd_Category, Fd_DE_ClfLiq
        )
        Dt_MEIn_ClfLiq = diet.calculate_Dt_MEIn_ClfLiq(
            Fd_ME_ClfLiq, Fd_DMIn_ClfLiq
        )
        Dt_DMIn_ClfLiq = Fd_DMIn_ClfLiq.sum()
        Dt_DMIn_ClfFor = Fd_DMIn_ClfFor.sum()
        Dt_DMIn_ClfStrt = diet.calculate_Dt_DMIn_ClfStrt(
            An_BW, Dt_MEIn_ClfLiq, Dt_DMIn_ClfLiq, Dt_DMIn_ClfFor,
            An_AgeDryFdStart, Env_TempCurr, DMIn_eqn, Trg_Dt_DMIn, coeff_dict
        )
        Dt_DMIn = calculate_Dt_DMIn_Calf1(
            Dt_DMIn_ClfLiq, Dt_DMIn_ClfStrt, Dt_DMIn_ClfFor
            )
    
    # Individual Heifer DMI Predictions
    elif DMIn_eqn in [2, 3, 4, 5, 6, 7]:
        Dt_DMIn_BW_LateGest_i = calculate_Dt_DMIn_BW_LateGest_i(
            An_PrePartWklim, Kb_LateGest_DMIn, coeff_dict
            )
        Dt_DMIn_Heif_LateGestInd = calculate_Dt_DMIn_Heif_LateGestInd(
            An_BW, Dt_DMIn_BW_LateGest_i
            )
        
        if DMIn_eqn == 2:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_NRCa(An_BW, An_BW_mature),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_NRCa(An_BW, An_BW_mature)
        
        if DMIn_eqn == 3:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_NRCad(An_BW, An_BW_mature, Dt_NDF),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_NRCad(
                    An_BW, An_BW_mature, Dt_NDF
                    )

        if DMIn_eqn == 4:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_H1(An_BW),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_H1(An_BW)

        if DMIn_eqn == 5:
            Dt_NDFdev_DMI = calculate_Dt_NDFdev_DMI(An_BW, Dt_NDF)
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_H2(An_BW, Dt_NDFdev_DMI),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_H2(An_BW, Dt_NDFdev_DMI)

        if DMIn_eqn == 6:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_HJ1(An_BW),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_HJ1(An_BW)
        
        if DMIn_eqn == 7:
            Dt_NDFdev_DMI = calculate_Dt_NDFdev_DMI(An_BW, Dt_NDF)
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_HJ2(An_BW, Dt_NDFdev_DMI),
                    Dt_DMIn_Heif_LateGestInd)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_HJ2(An_BW, Dt_NDFdev_DMI)

    elif DMIn_eqn == 8:
        Dt_DMIn = calculate_Dt_DMIn_Lact1(
            An_BW, An_BCS, An_LactDay, An_Parity_rl, Trg_NEmilkOut
            )

    elif DMIn_eqn == 9:
        Dt_DMIn = calculate_Dt_DMIn_Lact2(
            Dt_ForNDF, Dt_ADF, Dt_NDF, Dt_ForDNDF48_ForNDF, Trg_MilkProd
        )

    elif DMIn_eqn == 10:
        Dt_DMIn_BW_LateGest_i = calculate_Dt_DMIn_BW_LateGest_i(
            An_PrePartWklim, Kb_LateGest_DMIn, coeff_dict
            )
        Dt_DMIn_BW_LateGest_p = calculate_Dt_DMIn_BW_LateGest_p(
            An_PrePartWkDurat, Kb_LateGest_DMIn, coeff_dict
            )
        if An_PrePartWk > An_PrePartWkDurat:
            Dt_DMIn = min(
                calculate_Dt_DMIn_DryCow1_FarOff(An_BW, Dt_DMIn_BW_LateGest_i),
                calculate_Dt_DMIn_DryCow1_Close(An_BW, Dt_DMIn_BW_LateGest_p))
        else:
            Dt_DMIn = calculate_Dt_DMIn_DryCow1_FarOff(
                An_BW, Dt_DMIn_BW_LateGest_i
                )

    elif DMIn_eqn == 11:
        Dt_DMIn_DryCow_AdjGest = calculate_Dt_DMIn_DryCow_AdjGest(
            An_GestDay, An_GestLength, An_BW
        )
        Dt_DMIn = calculate_Dt_DMIn_DryCow2(
            An_BW, Dt_DMIn_DryCow_AdjGest
            )
    
    elif DMIn_eqn in [12, 13, 14, 15, 16, 17]:
        Dt_DMIn_BW_LateGest_p = calculate_Dt_DMIn_BW_LateGest_p(
            An_PrePartWkDurat, Kb_LateGest_DMIn, coeff_dict
            )
        Dt_DMIn_Heif_LateGestPen = calculate_Dt_DMIn_Heif_LateGestPen(
            An_BW, Dt_DMIn_BW_LateGest_p
            )
        
        if DMIn_eqn == 12:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_NRCa(An_BW, An_BW_mature),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_NRCa(An_BW, An_BW_mature)
        
        if DMIn_eqn == 13:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_NRCad(An_BW, An_BW_mature, Dt_NDF),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_NRCad(
                    An_BW, An_BW_mature, Dt_NDF
                    )
        
        if DMIn_eqn == 14:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_H1(An_BW),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_H1(An_BW)
        
        if DMIn_eqn == 15:
            Dt_NDFdev_DMI = calculate_Dt_NDFdev_DMI(An_BW, Dt_NDF)
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_H2(An_BW, Dt_NDFdev_DMI),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_H2(An_BW, Dt_NDFdev_DMI)
    
        if DMIn_eqn == 16:
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_HJ1(An_BW),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_HJ1(An_BW)
        
        if DMIn_eqn == 17:
            Dt_NDFdev_DMI = calculate_Dt_NDFdev_DMI(An_BW, Dt_NDF)
            if An_PrePartWk > An_PrePartWkDurat:
                Dt_DMIn = min(
                    calculate_Dt_DMIn_Heif_HJ2(An_BW, Dt_NDFdev_DMI),
                    Dt_DMIn_Heif_LateGestPen)
            else:
                Dt_DMIn = calculate_Dt_DMIn_Heif_HJ2(An_BW, Dt_NDFdev_DMI)
  
    else:
        raise ValueError(
            f"Invalid value for DMIn_eqn: {DMIn_eqn}. Must be between 0 and 17."
            )
    
    return Dt_DMIn
    


"""Rumen digestion calculations.

This module provides functions to estimate the digestibility of different 
dietary components in the rumen, including neutral detergent fiber (NDF) 
and starch, and their passage through the digestive system.
"""


def calculate_Rum_dcNDF(
    Dt_DMIn: float, 
    Dt_NDFIn: float, 
    Dt_StIn: float, 
    Dt_CPIn: float, 
    Dt_ADFIn: float,
    Dt_ForWet: float
) -> float:
    Rum_dcNDF = (-31.9 + 0.721 * Dt_NDFIn / 
                 Dt_DMIn * 100 - 0.247 * Dt_StIn / 
                 Dt_DMIn * 100 + 6.63 * Dt_CPIn / 
                 Dt_DMIn * 100 - 0.211 * (Dt_CPIn / Dt_DMIn * 100)**2 - 
                 0.387 * Dt_ADFIn / Dt_DMIn / (Dt_NDFIn / Dt_DMIn) * 100 - 
                 0.121 * Dt_ForWet + 1.51 * Dt_DMIn)

    if Rum_dcNDF < 0.1 or Rum_dcNDF is None:  # Line 984
        Rum_dcNDF = 0.1
    return Rum_dcNDF


def calculate_Rum_dcSt(
    Dt_DMIn: float, 
    Dt_ForNDF: float, 
    Dt_StIn: float, 
    Dt_ForWet: float
) -> float:
    Rum_dcSt = (70.6 - 1.45 * Dt_DMIn + 0.424 * Dt_ForNDF + 1.39 * Dt_StIn / 
                Dt_DMIn * 100 - 0.0219 * (Dt_StIn / Dt_DMIn * 100)**2 - 
                0.154 * Dt_ForWet)
    if Rum_dcSt < 0.1:  # Line 992
        Rum_dcSt = 0.1
    elif Rum_dcSt > 100:  # Line 993
        Rum_dcSt = 100
    return Rum_dcSt


def calculate_Rum_DigNDFIn(Rum_dcNDF: float, Dt_NDFIn: float) -> float:
    Rum_DigNDFIn = Rum_dcNDF / 100 * Dt_NDFIn
    return Rum_DigNDFIn


def calculate_Rum_DigStIn(Rum_dcSt: float, Dt_StIn: float) -> float:
    Rum_DigStIn = Rum_dcSt / 100 * Dt_StIn # Line 998
    return Rum_DigStIn


def calculate_Rum_DigNDFnfIn(Rum_dcNDF: float, Dt_NDFnfIn: float) -> float:
    """
    Rum_DigNDFnfIn: Rumen digested Nitrogen free NDF intake, kg
    """
    Rum_DigNDFnfIn = Rum_dcNDF / 100 * Dt_NDFnfIn  
    # Not used.  IF used, should consider infusions, Line 996
    return Rum_DigNDFnfIn


def calculate_Du_StPas(
    Dt_StIn: float, 
    InfRum_StIn: float,
    Rum_DigStIn: float
) -> float:
    """
    Du_StPas: Duodenal starch passage?, kg?
    """
    Du_StPas = Dt_StIn + InfRum_StIn - Rum_DigStIn  # Line 999
    if Du_StPas < 0:  
        # All grass diets predicted to be very slightly negative, Line 1000
        Du_StPas = 0
    return Du_StPas


def calculate_Du_NDFPas(
    Dt_NDFIn: float, 
    Inf_NDFIn: float,
    Rum_DigNDFIn: float
) -> float:
    """
    Du_NDFPas: Duodenal NDF passage?, kg?
    """
    Du_NDFPas = Dt_NDFIn + Inf_NDFIn - Rum_DigNDFIn  # Line 1001
    return Du_NDFPas



"""Water intake and balance calculations.

This module provides functions to estimate voluntary water intake, 
water loss through manure, and water balance, adjusted for diet and 
environmental conditions.
"""


def calculate_An_WaIn(
    An_StatePhys: str, 
    Dt_DMIn: float, 
    Dt_DM: float,
    Dt_Na: float, 
    Dt_K: float, 
    Dt_CP: float,
    Env_TempCurr: float
) -> float:
    """
    An_WaIn: predicted voluntary water intake, kg/d

    From Appuhamy et al., 2016.  Requires physiological state and mean daily ambient temp.
    Based on Diet DMI.  Should perhaps add infusions, but no minerals or DM specified for infusions?
    """
    if An_StatePhys == "Lactating Cow":  # Line 966
        An_WaIn = (-91.1 + 2.93 * Dt_DMIn + 0.61 * Dt_DM + 
                   0.062 * (Dt_Na / 0.023 + Dt_K / 0.039) * 10 + 
                   2.49 * Dt_CP + 0.76 * Env_TempCurr)  # Line 963-964
        # Low DMI, CP, and Na results in too low of WaIn of 10 kg/d.
        # Consider trapping values below 22 which is the min from observed 
        # data. MDH from RM.
    # elif An_StatePhys == "Heifer":  # Line 967
    elif An_StatePhys in ["Heifer", "Dry Cow"]:
        An_WaIn = (1.16 * Dt_DMIn + 0.23 * Dt_DM + 0.44 * Env_TempCurr + 
                   0.061 * (Env_TempCurr - 16.4)**2) # Line 965
    else:  # Line 968
        An_WaIn = None  
    #the above (An_Wa_In_Lact/Dry) does not apply to calves/other thus set to NA
    return An_WaIn


def calculate_An_Wa_Insens(
    An_WaIn: float, 
    Mlk_Prod: float,
    Man_Wa_out: float
) -> float:
    """
    An_Wa_Insens: Water baalance? (L/d)
    """
    An_Wa_Insens = An_WaIn - Mlk_Prod - Man_Wa_out  
    # L/d; by difference, Line 3333
    return An_Wa_Insens


def calculate_WaIn_Milk(An_WaIn: float, Mlk_Prod: float) -> float:
    """
    WaIn_Milk: Water intake per kg milk production (L/kg)
    """
    WaIn_Milk = An_WaIn / Mlk_Prod  # L/kg, Line 3334
    return WaIn_Milk



"""Functions to calculate values related to the body composition of animals.

This includes fat, protein, and water content. There are also estimates of 
composition change based on the change in body reserves.
"""

import math

import numpy as np


def calculate_Frm_Gain_empty(
    Frm_Gain: float, 
    Dt_DMIn_ClfLiq: float, 
    Dt_DMIn_ClfStrt: float,
    An_GutFill_BW: float
) -> float:
    Frm_Gain_empty = Frm_Gain * (1 - An_GutFill_BW)  
    # Line 2439, Assume the same gut fill for frame gain
    if Dt_DMIn_ClfLiq > 0 and Dt_DMIn_ClfStrt > 0:
        # slightly different for grain & milk fed, Line 2440
        Frm_Gain_empty = Frm_Gain * 0.91
    return Frm_Gain_empty


def calculate_Body_Gain_empty(
    Frm_Gain_empty: float, 
    Rsrv_Gain_empty: float
) -> float:
    Body_Gain_empty = Frm_Gain_empty + Rsrv_Gain_empty  # Line 2442
    return Body_Gain_empty


def calculate_NPGain_RsrvGain(coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"CPGain_RsrvGain": 0.068, "Body_NP_CP": 0.86}
    
    calculate_NPGain_RsrvGain(coeff_dict = coeff_dict)
    ```
    """
    NPGain_RsrvGain = coeff_dict['CPGain_RsrvGain'] * coeff_dict['Body_NP_CP']  
    # Line 2467
    return NPGain_RsrvGain


def calculate_Rsrv_NPgain(
    NPGain_RsrvGain: float, 
    Rsrv_Gain_empty: float
) -> float:
    Rsrv_NPgain = NPGain_RsrvGain * Rsrv_Gain_empty  # Line 2468
    return Rsrv_NPgain


def calculate_Body_NPgain(Frm_NPgain: float, Rsrv_NPgain: float) -> float:
    Body_NPgain = Frm_NPgain + Rsrv_NPgain  # Line 2473
    return Body_NPgain


def calculate_Body_CPgain(Body_NPgain: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.86}
    
    calculate_Body_CPgain(
        Body_NPgain = 50.0, coeff_dict = coeff_dict
    )
    ```
    """
    Body_CPgain = Body_NPgain / coeff_dict['Body_NP_CP']  # Line 2475
    return Body_CPgain


def calculate_Body_CPgain_g(Body_CPgain: float) -> float:
    Body_CPgain_g = Body_CPgain * 1000  # Line 2477
    return Body_CPgain_g


def calculate_Rsrv_Gain(Trg_RsrvGain: float) -> float:
    """
    Trg_RsrvGain (float): Target gain or loss in body reserves 
                          (66% fat, 8% CP) in kg fresh weight/day
    """
    Rsrv_Gain = Trg_RsrvGain  # Line 2435
    return Rsrv_Gain


def calculate_Rsrv_Gain_empty(Rsrv_Gain: float) -> float:
    """
    Rsrv_Gain_empty: Body reserve gain assuming no gut fill association, kg/d
    """
    Rsrv_Gain_empty = Rsrv_Gain  
    # Assume no gut fill associated with reserves gain, Line 2441
    return Rsrv_Gain_empty


def calculate_Rsrv_Fatgain(Rsrv_Gain_empty: float, coeff_dict: dict) -> float:
    """
    FatGain_RsrvGain: Conversion factor from reserve gain to fat gain
    Rsrv_Fatgain: Body reserve fat gain kg/d
    
    Examples
    --------
    ```
    coeff_dict = {"FatGain_RsrvGain": 0.622}
    
    calculate_Rsrv_Fatgain(
        Rsrv_Gain_empty = 40.0, coeff_dict = coeff_dict
    )
    ```
    """
    Rsrv_Fatgain = coeff_dict['FatGain_RsrvGain'] * Rsrv_Gain_empty  # Line 2453
    return Rsrv_Fatgain


def calculate_CPGain_FrmGain(An_BW: float, An_BW_mature: float) -> float:
    """
    CPGain_FrmGain: CP gain per unit of frame gain kg/d
    """
    CPGain_FrmGain = 0.201 - 0.081 * An_BW / An_BW_mature  
    # CP gain / gain for heifers, Line 2458
    return CPGain_FrmGain


def calculate_Rsrv_CPgain(
    CPGain_FrmGain: float,
    Rsrv_Gain_empty: float
) -> float:
    """
    Rsrv_CPgain: CP portion of body reserve gain, g/g
    """
    Rsrv_CPgain = CPGain_FrmGain * Rsrv_Gain_empty  # Line 2470
    return Rsrv_CPgain


def calculate_FatGain_FrmGain(
    An_StatePhys: str, 
    An_REgain_Calf: float,
    An_BW: float, 
    An_BW_mature: float
) -> float:
    """
    FatGain_FrmGain: Fat gain per unit frame gain, kg/kg EBW (Empty body weight)
    This is the proportion of the empty body weight that is fat, which increases 
    (and protein decreases) as the animal matures
    This is why it is scaled to a proportion of mature BW (An_BW / An_BW_mature)
    Also in equation 20-253
    """
    if An_StatePhys == "Calf":
        FatGain_FrmGain = 0.0786 + 0.0370 * An_REgain_Calf # Calves,..., g/g EBW
    else:
        FatGain_FrmGain = (0.067 + 0.375 * An_BW / An_BW_mature)
    if math.isnan(FatGain_FrmGain):
        FatGain_FrmGain = 0
    return FatGain_FrmGain


def calculate_Frm_Gain(Trg_FrmGain: float) -> float:
    """
    Frame gain, kg/d
    """
    Frm_Gain = Trg_FrmGain  
    # Add any predictions of ADG and select Trg or Pred ADG here, Line 2434
    return Frm_Gain


def calculate_Frm_Fatgain(
    FatGain_FrmGain: float,
    Frm_Gain_empty: float
) -> float:
    """
    Frm_Fatgain: Frame fat gain kg/?
    FatGain_FrmGain is the kg of fat per kg of empty frame gain, based on the BW of the animal as a % of mature BW. 
    Here, the actual kg/kg * EBG (kg/d) = frame fat gain kg/d
    In the book, the Fat_ADG from equation 3-20a (which is  FatGain_FrmGain here) is corrected for empty body weight (e.g. x 0.85), keeping it in kg/kg units. 
     this is a mix of Equations 11-5a and 11-6a (but 11-6a assumes 0.85, not EBG/ADG)
    """
    Frm_Fatgain = FatGain_FrmGain * Frm_Gain_empty  # Line 2452
    return Frm_Fatgain


def calculate_NPGain_FrmGain(CPGain_FrmGain: float, coeff_dict: dict) -> float:
    """
    NPGain_FrmGain: Net protein gain per unit frame gain
    NOTE for these gain per unit gain values I believe they are unitless/have units g/g, not much said in the R code
    
    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.86}
    
    calculate_NPGain_FrmGain(
        CPGain_FrmGain = 60.0, coeff_dict = coeff_dict
    )
    ```
    """
    # Convert to CP to TP gain / gain, Line 2459
    NPGain_FrmGain = CPGain_FrmGain * coeff_dict['Body_NP_CP']
    return NPGain_FrmGain


def calculate_Frm_NPgain(
    An_StatePhys: str, 
    NPGain_FrmGain: float,
    Frm_Gain_empty: float, 
    Body_Gain_empty: float,
    An_REgain_Calf: float
) -> float:
    """
    Frm_NPgain: NP portion of frame gain
    """
    Frm_NPgain = NPGain_FrmGain * Frm_Gain_empty  # TP gain, Line 2460
    if An_StatePhys == "Calf":
        Frm_NPgain = (166.22 * Body_Gain_empty + 6.13 * An_REgain_Calf /
                      Body_Gain_empty) / 1000  # Line 2461
    return Frm_NPgain


def calculate_Frm_CPgain(Frm_NPgain: float, coeff_dict: dict) -> float:
    """
    Frm_CPgain: CP portion of frame gain

    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.86}
    
    calculate_Frm_CPgain(
        Frm_NPgain = 50.0, coeff_dict = coeff_dict
    )
    ```
    """
    Frm_CPgain = Frm_NPgain / coeff_dict['Body_NP_CP']  # Line 2463
    return Frm_CPgain


def calculate_Body_NPgain_g(Body_NPgain: float) -> float:
    """
    Body_NPgain_g: Net protein from frame and reserve gain, g
    """
    Body_NPgain_g = Body_NPgain * 1000  # Line 2475
    return Body_NPgain_g


def calculate_An_BWmature_empty(
    An_BW_mature: float, 
    coeff_dict: dict
) -> float:
    """
    An_BWmature_empty: kg bodyweight with no gut fill 

    Examples
    --------
    ```
    coeff_dict = {"An_GutFill_BWmature": 0.18}
    
    calculate_An_BWmature_empty(
        An_BW_mature = 600.0, coeff_dict = coeff_dict
    )
    ```
    """
    An_BWmature_empty = An_BW_mature * (1 - coeff_dict['An_GutFill_BWmature'])
    return An_BWmature_empty


def calculate_Body_Gain(Frm_Gain: float, Rsrv_Gain: float) -> float:
    """
    Body_Gain: Bodyweight gain, kg/d
    """
    Body_Gain = Frm_Gain + Rsrv_Gain  # Line 2436
    return Body_Gain


def calculate_Trg_BWgain(Trg_FrmGain: float, Trg_RsrvGain: float) -> float:
    """
    Trg_BWgain: Target rate of bodyweight gain, kg/d
    """
    Trg_BWgain = Trg_FrmGain + Trg_RsrvGain  
    # This could also be generated by prediction equations, Line 224
    return Trg_BWgain


def calculate_Trg_BWgain_g(Trg_BWgain: float) -> float:
    """
    Trg_BWgain_g: Target rate of bodyweight gain, g/d
    """
    Trg_BWgain_g = Trg_BWgain * 1000  # Line 225
    return Trg_BWgain_g


def calculate_Conc_BWgain(GrUter_BWgain: float, Uter_BWgain: float) -> float:
    """
    Conc_BWgain: ?
    """
    Conc_BWgain = GrUter_BWgain - Uter_BWgain  # Line 2349
    return Conc_BWgain


def calculate_BW_BCS(An_BW: float) -> float:
    """
    BW_BCS: Calculate BCS from BW?
    """
    BW_BCS = 0.094 * An_BW  
    # Each BCS represents 94 g of weight per kg of BW, Line 2395
    return BW_BCS


def calculate_An_BWnp3(An_BWnp: float, An_BCS: float) -> float:
    """
    An_BWnp3: Non-pregnant BW standardized to BCS 3(kg)
    """
    An_BWnp3 = An_BWnp / (
        1 + 0.094 * (An_BCS - 3)
    )  # BWnp standardized to BCS of 3 using 9.4% of BW/unit of BCS, Line 2397
    return An_BWnp3


def calculate_An_GutFill_Wt_Erdman(
    Dt_DMIn: float, 
    InfRum_DMIn: float,
    InfSI_DMIn: float
) -> float:
    """
    An_GutFill_Wt_Erdman: Gut fill from Erdman et al. 2017
    """
    An_GutFill_Wt_Erdman = 5.9 * (Dt_DMIn + InfRum_DMIn + InfSI_DMIn)  
    # cows only, from Erdman et al. 2017; not used, Line 2411
    return An_GutFill_Wt_Erdman


def calculate_An_BWnp_empty(An_BWnp: float, An_GutFill_Wt: float) -> float:
    """
    An_BWnp_empty: Non pregnant empty BW (kg). 
    Comment says 3 Std BCS standardized but this does not use An_BWnp3?
    """
    An_BWnp_empty = An_BWnp - An_GutFill_Wt  # BCS3 Std EBW, Line 2416
    return An_BWnp_empty


def calculate_An_BWnp3_empty(An_BWnp3: float, An_GutFill_Wt: float) -> float:
    """
    An_BWnp3_empty: Non pregnant empty bodyweight standardized 3 Std
    """
    An_BWnp3_empty = An_BWnp3 - An_GutFill_Wt  # BCS3 Std EBWnp, Line 2417
    return An_BWnp3_empty


def calculate_Body_Fat_EBW(An_BW: float, An_BW_mature: float) -> float:
    """
    Body_Fat_EBW: g fat / g empty BW
    """
    Body_Fat_EBW = 0.067 + 0.188 * An_BW / An_BW_mature  # Line 2420
    return Body_Fat_EBW


def calculate_Body_NonFat_EBW(Body_Fat_EBW: float) -> float:
    """
    Body_NonFat_EBW: Non fat empty bodyweight (g/g EBW)
    """
    Body_NonFat_EBW = 1 - Body_Fat_EBW  # Line 2421
    return Body_NonFat_EBW


def calculate_Body_CP_EBW(Body_NonFat_EBW: float) -> float:
    """
    Body_CP_EBW: g CP / g EBW
    """
    Body_CP_EBW = 0.215 * Body_NonFat_EBW  # Line 2422
    return Body_CP_EBW


def calculate_Body_Ash_EBW(Body_NonFat_EBW: float) -> float:
    """
    Body_Ash_EBW: g Ash / g EBW
    """
    Body_Ash_EBW = 0.056 * Body_NonFat_EBW  # Line 2423
    return Body_Ash_EBW


def calculate_Body_Wat_EBW(Body_NonFat_EBW: float) -> float:
    """
    Body_Wat_EBW: g water? / g EBW
    """
    Body_Wat_EBW = 0.729 * Body_NonFat_EBW  # Line 2424
    return Body_Wat_EBW


def calculate_Body_Fat(An_BWnp_empty: float, Body_Fat_EBW: float) -> float:
    """
    Body_Fat: Body fat (g)
    """
    Body_Fat = An_BWnp_empty * Body_Fat_EBW  
    # Using a non-pregnant BW, Line 2426
    return Body_Fat


def calculate_Body_NonFat(
    An_BWnp_empty: float,
    Body_NonFat_EBW: float
) -> float:
    """
    Body_NonFat: Non fat bodyweight (g)
    """
    Body_NonFat = An_BWnp_empty * Body_NonFat_EBW  # Line 2427
    return Body_NonFat


def calculate_Body_CP(An_BWnp_empty: float, Body_NonFat_EBW: float) -> float:
    """
    Body_CP: CP bodyweight (g)
    """
    Body_CP = An_BWnp_empty * Body_NonFat_EBW  # Line 2428
    return Body_CP


def calculate_Body_Ash(An_BWnp_empty: float, Body_Ash_EBW: float) -> float:
    """
    Body_Ash: Ash bodyweight (g)
    """
    Body_Ash = An_BWnp_empty * Body_Ash_EBW  # Line 2429
    return Body_Ash


def calculate_Body_Wat(An_BWnp_empty: float, Body_Wat_EBW: float) -> float:
    """
    Body_Wat: Water bodyweight (g)
    """
    Body_Wat = An_BWnp_empty * Body_Wat_EBW  # Line 2430
    return Body_Wat


def calculate_An_BodConcgain(Body_Gain: float, Conc_BWgain: float) -> float:
    """
    An_BodConcgain: ? (kg/d)
    """
    An_BodConcgain = Body_Gain + Conc_BWgain  # Minus fetal fluid, Line 2437
    return An_BodConcgain


def calculate_NonFatGain_FrmGain(FatGain_FrmGain: float) -> float:
    """
    NonFatGain_FrmGain: Non fat frame gain (g/g)
    """
    NonFatGain_FrmGain = 1 - FatGain_FrmGain  # Line 2450
    return NonFatGain_FrmGain


def calculate_Body_Fatgain(Frm_Fatgain: float, Rsrv_Fatgain: float) -> float:
    """
    Body_Fatgain: Fat gain (g/g)
    """
    Body_Fatgain = Frm_Fatgain + Rsrv_Fatgain  # Line 2454
    return Body_Fatgain


def calculate_Body_NonFatGain(
    Body_Gain_empty: float,
    Body_Fatgain: float
) -> float:
    """
    Body_NonFatGain: Non fat gain (g/g)
    """
    Body_NonFatGain = Body_Gain_empty - Body_Fatgain  # Line 2455
    return Body_NonFatGain


def calculate_Frm_CPgain_g(Frm_CPgain: float) -> float:
    """
    Frm_CPgain_g: Body frame crude protein gain (kg)
    """
    Frm_CPgain_g = Frm_CPgain * 1000  # Line 2464
    return Frm_CPgain_g


def calculate_Rsrv_CPgain_g(Rsrv_CPgain: float) -> float:
    """
    Rsrv_CPgain_g: Body reserve crude protein gain (g/d)
    """
    Rsrv_CPgain_g = Rsrv_CPgain * 1000  # Line 2471
    return Rsrv_CPgain_g


def calculate_Body_AshGain(Body_NonFatGain: float) -> float:
    """
    Body_AshGain: Body ash gain (g/g)
    """
    Body_AshGain = 0.056 * Body_NonFatGain  
    # Alternative method of estimation Body and Frm Ash, Line 2484
    return Body_AshGain


def calculate_Frm_AshGain(Body_AshGain: float) -> float:
    """
    Frm_AshGain: Body frame ash gain (g/g)
    """
    Frm_AshGain = Body_AshGain  # Line 2485
    return Frm_AshGain


def calculate_WatGain_RsrvGain(
    NPGain_RsrvGain: float,
    coeff_dict: dict
) -> float:
    """
    WatGain_RsrvGain: Body reserve water gain (g/g)
    
    Examples
    --------
    ```
    coeff_dict = {"FatGain_RsrvGain": 0.622, "AshGain_RsrvGain": 0.02}
    
    calculate_WatGain_RsrvGain(
        NPGain_RsrvGain = 0.20, coeff_dict = coeff_dict
    )
    ```
    """
    WatGain_RsrvGain = (100 - coeff_dict['FatGain_RsrvGain'] - NPGain_RsrvGain - 
                        coeff_dict['AshGain_RsrvGain'])  # Line 2489
    return WatGain_RsrvGain


def calculate_Rsrv_WatGain(
    WatGain_RsrvGain: float,
    Rsrv_Gain_empty: float
) -> float:
    """
    Rsrv_WatGain: Body reserve water gain (g/g)
    """
    Rsrv_WatGain = WatGain_RsrvGain * Rsrv_Gain_empty  # Line 2491
    return Rsrv_WatGain


def calculate_Body_WatGain(Body_NonFatGain: float) -> float:
    """
    Body_WatGain: Bodyweight water gain (g/g)
    """
    Body_WatGain = 0.729 * Body_NonFatGain  
    # Alternative method of estimation, Line 2493
    return Body_WatGain


def calculate_Frm_WatGain(Body_WatGain: float, Rsrv_WatGain: float) -> float:
    """
    Frm_WatGain: Body frame water gain (g/g)
    """
    Frm_WatGain = Body_WatGain - Rsrv_WatGain  # Line 2494
    return Frm_WatGain


def calculate_An_MPavail_Gain_Trg(
    An_MPIn: float, 
    An_MPuse_g_Trg: float,
    Body_MPUse_g_Trg: float
) -> float:
    """
    An_MPavail_Gain_Trg: MP available for target gain? (kg/d)
    """
    An_MPavail_Gain_Trg = (An_MPIn - An_MPuse_g_Trg / 
                           1000 + Body_MPUse_g_Trg / 1000)  # Line 2712
    return An_MPavail_Gain_Trg


def calculate_Body_NPgain_MPalowTrg_g(
    An_MPavail_Gain_Trg: float,
    Kg_MP_NP_Trg: float
) -> float:
    """
    Body_NPgain_MPalowTrg_g: MP allowable NP gain (g/d)
    """
    Body_NPgain_MPalowTrg_g = An_MPavail_Gain_Trg * Kg_MP_NP_Trg * 1000  
    # g NP gain/d, Line 2713
    return Body_NPgain_MPalowTrg_g


def calculate_Body_CPgain_MPalowTrg_g(
    Body_NPgain_MPalowTrg_g: float,
    coeff_dict: dict
) -> float:
    """
    Body_CPgain_MPalowTrg_g: MP allowable CP gain (g/d)
    
    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.80}
    
    calculate_Body_CPgain_MPalowTrg_g(
        Body_NPgain_MPalowTrg_g = 40.0, 
        coeff_dict = coeff_dict
    )
    ```
    """
    Body_CPgain_MPalowTrg_g = Body_NPgain_MPalowTrg_g / coeff_dict['Body_NP_CP']  
    # g CP gain/d, Line 2714
    return Body_CPgain_MPalowTrg_g


def calculate_Body_Gain_MPalowTrg_g(
    Body_NPgain_MPalowTrg_g: float,
    NPGain_FrmGain: float
) -> float:
    """
    Body_Gain_MPalowTrg_g: MP allowable body gain (g/d)
    """
    Body_Gain_MPalowTrg_g = Body_NPgain_MPalowTrg_g / NPGain_FrmGain  
    # g/d, Assume all is frame gain, Line 2715
    return Body_Gain_MPalowTrg_g


def calculate_Body_Gain_MPalowTrg(Body_Gain_MPalowTrg_g: float) -> float:
    """
    Body_Gain_MPalowTrg: MP allowable body gain (kg/d)
    """
    Body_Gain_MPalowTrg = Body_Gain_MPalowTrg_g / 1000  # Line 2716
    return Body_Gain_MPalowTrg


def calculate_An_MEavail_Grw(
    An_MEIn: float, 
    An_MEmUse: float,
    Gest_MEuse: float, 
    Mlk_MEout: float
) -> float:
    """
    An_MEavail_Grw: ME available for growth (Mcal/d)
    """
    An_MEavail_Grw = An_MEIn - An_MEmUse - Gest_MEuse - Mlk_MEout  # Line 2947
    return An_MEavail_Grw


def calculate_Kg_ME_NE(
    Frm_NEgain: float, 
    Rsrv_NEgain: float, 
    Kr_ME_RE: float,
    Kf_ME_RE: float
) -> float:
    """
    Kg_ME_NE: ME to NE for NE allowable gain?
    """
    if Frm_NEgain + Rsrv_NEgain == 0:
        Kg_ME_NE = 0
    else:
    #Use a weighted average of Kf and Kr to predict allowable gain at that mix of Frm and Rsrv gain.
        Kg_ME_NE = (Kf_ME_RE * Frm_NEgain / 
                    (Frm_NEgain + Rsrv_NEgain) + Kr_ME_RE * Rsrv_NEgain / 
                    (Frm_NEgain + Rsrv_NEgain))
    return Kg_ME_NE


def calculate_Body_Gain_NEalow(
    An_MEavail_Grw: float, 
    Kg_ME_NE: float,
    Body_NEgain_BWgain: float
) -> float:
    """
    Body_Gain_NEalow: NE allowable body gain
    """
    if (Body_NEgain_BWgain != 0 # Line 2950
        and not np.isnan(An_MEavail_Grw) 
        and not np.isnan(Kg_ME_NE) 
        and not np.isnan(Body_NEgain_BWgain)
        ):
        Body_Gain_NEalow = An_MEavail_Grw * Kg_ME_NE / Body_NEgain_BWgain 
    else:
        Body_Gain_NEalow = np.nan
    return Body_Gain_NEalow


def calculate_An_BodConcgain_NEalow(
    Body_Gain_NEalow: float,
    Conc_BWgain: float
) -> float:
    """
    An_BodConcgain_NEalow: ?
    """
    An_BodConcgain_NEalow = Body_Gain_NEalow + Conc_BWgain  # Line 2954
    return An_BodConcgain_NEalow


def calculate_Body_Fatgain_NEalow(Body_Gain_NEalow: float) -> float:
    """
    Body_Fatgain_NEalow: NE allowable body fat gain (kg/d?)
    """
    Body_Fatgain_NEalow = 0.85 * (Body_Gain_NEalow / 0.85 -1.19) / 8.21  
    # Line 2955
    return Body_Fatgain_NEalow


def calculate_Body_NPgain_NEalow(Body_Fatgain_NEalow: float) -> float:
    """
    Body_NPgain_NEalow: NE allowable body NP gain (kg/d?)
    """
    Body_NPgain_NEalow = 0.85 * (1 - Body_Fatgain_NEalow / 0.85) * 0.215
    return Body_NPgain_NEalow


def calculate_An_Days_BCSdelta1(
    BW_BCS: float,
    Body_Gain_NEalow: float
) -> float:
    """
    An_Days_BCSdelta1: Days to gain/lose 1 BCS 
    """
    An_Days_BCSdelta1 = BW_BCS / Body_Gain_NEalow  
    # days to gain or lose 1 BCS (9.4% of BW), 5 pt scale., Line 2958
    return An_Days_BCSdelta1


def calculate_Rsrv_AshGain(
    Rsrv_Gain_empty: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"AshGain_RsrvGain": 0.02}
    
    calculate_Rsrv_AshGain(
        Rsrv_Gain_empty = 40.0, coeff_dict = coeff_dict
    )
    ```
    """
    Rsrv_AshGain = coeff_dict["AshGain_RsrvGain"] * Rsrv_Gain_empty   # Line 2481
    return Rsrv_AshGain



"""Gestation-related calculations for fetal growth and uterine changes.

This module covers the estimation of fetal growth, uterine weight changes, and 
nutrient requirements during pregnancy.
"""

import math

import numpy as np
import pandas as pd


def calculate_Uter_Wtpart(Fet_BWbrth: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"UterWt_FetBWbrth": 1.816}
    
    calculate_Uter_Wtpart(
        Fet_BWbrth = 30.0, coeff_dict = coeff_dict
    )
    ```
    """
    Uter_Wtpart = Fet_BWbrth * coeff_dict['UterWt_FetBWbrth']
    return Uter_Wtpart


def calculate_Uter_Wt(
    An_Parity_rl: int, 
    An_AgeDay: int, 
    An_LactDay: int, 
    An_GestDay: int,
    An_GestLength: int, 
    Uter_Wtpart: float, 
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {
        "Uter_Wt": 0.204, "Uter_Ksyn": 2.42e-2, "Uter_KsynDecay": 3.53e-5, 
        "Uter_Kdeg": 0.20
    }
    
    calculate_Uter_Wt(
        An_Parity_rl = 1, An_AgeDay = 250, An_LactDay = 50, An_GestDay = 150, 
        An_GestLength = 280, Uter_Wtpart = 45.0, coeff_dict = coeff_dict
    )
    ```
    """
    Uter_Wt = coeff_dict['Uter_Wt_coeff']  # Line 2312
    if An_AgeDay < 240:  # Line 2313
        Uter_Wt = 0

    if An_GestDay > 0 and An_GestDay <= An_GestLength:  # Line 2314
        Uter_Wt = (Uter_Wtpart * 
                   math.exp(-(coeff_dict['Uter_Ksyn'] - 
                              coeff_dict['Uter_KsynDecay'] * An_GestDay) * 
                              (An_GestLength - An_GestDay)))
    
    if An_GestDay <= 0 and An_LactDay > 0 and An_LactDay < 100:  # Line 2316
        Uter_Wt = ((Uter_Wtpart-0.204) * 
                   math.exp(-coeff_dict['Uter_Kdeg'] * An_LactDay) + 0.204)

    if An_Parity_rl > 0 and Uter_Wt < 0.204:  # Line 2318
        Uter_Wt = 0.204
    return Uter_Wt


def calculate_GrUter_Wtpart(Fet_BWbrth: float, coeff_dict: dict) -> float:
    """
    GrUterWt_FetBWbrth default 1.816 based on equation 20-225
    See Equation 3-15a, supposed to be 1.825?
    
    Examples
    --------
    ```
    coeff_dict = {"GrUterWt_FetBWbrth": 1.816}
    
    calculate_GrUter_Wtpart(
        Fet_BWbrth = 30.0, coeff_dict = coeff_dict
    )
    ```
    """
    GrUter_Wtpart = Fet_BWbrth * coeff_dict['GrUterWt_FetBWbrth'] # Line 2322
    return GrUter_Wtpart


def calculate_GrUter_Wt(
    An_GestDay: int, 
    An_GestLength: int, 
    Uter_Wt: float, 
    GrUter_Wtpart: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {
        "GrUter_Ksyn": 2.43e-2, "GrUter_KsynDecay": 2.45e-5
    }
    
    calculate_GrUter_Wt(
        An_GestDay = 150, An_GestLength = 280, Uter_Wt = 40.0, 
        GrUter_Wtpart = 55.0, coeff_dict = coeff_dict
    )
    ```
    """
    GrUter_Wt = Uter_Wt
    if An_GestDay > 0 and An_GestDay <= An_GestLength:  # Line 2323-2327
        GrUter_Wt = (GrUter_Wtpart * 
                     math.exp(-(coeff_dict['GrUter_Ksyn'] - 
                                coeff_dict['GrUter_KsynDecay'] * An_GestDay) * 
                                (An_GestLength - An_GestDay)))
    if GrUter_Wt < Uter_Wt:
        GrUter_Wt = Uter_Wt
    return GrUter_Wt


def calculate_Uter_BWgain(
    An_LactDay: int, 
    An_GestDay: int, 
    An_GestLength: int, 
    Uter_Wt: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {
        "Uter_BWgain": 0, "Uter_Ksyn": 2.42e-2, "Uter_KsynDecay": 3.53e-5, 
        "Uter_Kdeg": 0.20
    }
    
    calculate_Uter_BWgain(
        An_LactDay = 50, An_GestDay = 150, An_GestLength = 280, Uter_Wt = 45.0, 
        coeff_dict = coeff_dict
    )
    ```
    """
    Uter_BWgain = coeff_dict['Uter_BWgain_coeff']
    if An_GestDay > 0 and An_GestDay <= An_GestLength:
        Uter_BWgain = (coeff_dict['Uter_Ksyn'] -
                       coeff_dict['Uter_KsynDecay'] * An_GestDay) * Uter_Wt
    if An_GestDay <= 0 and An_LactDay > 0 and An_LactDay < 100: 
        # uterine involution after calving
        Uter_BWgain = -coeff_dict['Uter_Kdeg'] * Uter_Wt
    return Uter_BWgain


def calculate_GrUter_BWgain(
    An_LactDay: int, 
    An_GestDay: int, 
    An_GestLength: int, 
    GrUter_Wt: float,
    Uter_BWgain: float, 
    coeff_dict: dict
) -> float:
    """
    Equation 3-17a
    
    Examples
    --------
    ```
    coeff_dict = {
        "GrUter_BWgain": 0, "GrUter_Ksyn": 2.43e-2, "GrUter_KsynDecay": 2.45e-5
    }
    
    calculate_GrUter_BWgain(
        An_LactDay = 50, An_GestDay = 150, An_GestLength = 280, 
        GrUter_Wt = 50.0, Uter_BWgain = 0.5, coeff_dict = coeff_dict
    )
    ```
    """
    GrUter_BWgain = coeff_dict['GrUter_BWgain_coeff']  # Line 2341-2345
    if An_GestDay > 0 and An_GestDay <= An_GestLength:
        GrUter_BWgain = (
            coeff_dict['GrUter_Ksyn'] -
            coeff_dict['GrUter_KsynDecay'] * An_GestDay) * GrUter_Wt
    if An_GestDay <= 0 and An_LactDay > 0 and An_LactDay < 100:
        GrUter_BWgain = Uter_BWgain
    return GrUter_BWgain


def calculate_Gest_NCPgain_g(
    GrUter_BWgain: float, 
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"CP_GrUtWt": 0.123}
    
    calculate_Gest_NCPgain_g(
        GrUter_BWgain = 50.0, coeff_dict = coeff_dict
    )
    ```
    """
    Gest_NCPgain_g = GrUter_BWgain * coeff_dict['CP_GrUtWt'] * 1000
    return Gest_NCPgain_g


def calculate_Gest_NPgain_g(Gest_NCPgain_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.86}
    
    calculate_Gest_NPgain_g(
        Gest_NCPgain_g = 2.5, coeff_dict = coeff_dict
    )
    ```
    """
    Gest_NPgain_g = Gest_NCPgain_g * coeff_dict['Body_NP_CP']
    return Gest_NPgain_g


def calculate_Gest_NPuse_g(Gest_NPgain_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Gest_NPother_g": 0}
    
    calculate_Gest_NPuse_g(
        Gest_NPgain_g = 2.0, coeff_dict = coeff_dict
    )
    ```
    """
    Gest_NPuse_g = Gest_NPgain_g + coeff_dict['Gest_NPother_g']  # Line 2366
    return Gest_NPuse_g


def calculate_Gest_CPuse_g(Gest_NPuse_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"Body_NP_CP": 0.86}
    
    calculate_Gest_CPuse_g(
        Gest_NPuse_g = 2.1, coeff_dict = coeff_dict
    )
    ```
    """
    Gest_CPuse_g = Gest_NPuse_g / coeff_dict['Body_NP_CP']  # Line 2367
    return Gest_CPuse_g


def calculate_An_PostPartDay(An_LactDay: int) -> int:
    """
    An_LactDay: Day of lactation
    An_PostPartDay: Days postpartum 
    """
    # Calculate day postpartum and trap nonsense values
    if An_LactDay <= 0:  # Line 220
        An_PostPartDay = 0
    elif An_LactDay > 100:  # Line 221
        An_PostPartDay = 100
    else:
        An_PostPartDay = An_LactDay
    return An_PostPartDay


def calculate_An_Preg(An_GestDay: int, An_GestLength: int) -> int:
    """
    An_Preg: 0 if not pregnant, 1 if pregnant
    """
    #Create a pregnancy switch for use in calculations.
    if (An_GestDay > 0) & (An_GestDay <= An_GestLength):  # Line 2293
        An_Preg = 1
    else:
        An_Preg = 0
    return An_Preg


def calculate_Fet_Wt(
    An_GestDay: int, 
    An_GestLength: int, 
    Fet_BWbrth: float,
    coeff_dict: dict
) -> float:
    """
    Fet_Wt: Fetal weight at any time (kg)
    
    Examples
    --------
    ```
    coeff_dict = {"Fet_Ksyn": 5.16e-2, "Fet_KsynDecay": 7.59e-5}
    
    calculate_Fet_Wt(
        An_GestDay = 150, An_GestLength = 280, Fet_BWbrth = 40.0, 
        coeff_dict = coeff_dict
    )
    ```
    """
    if (An_GestDay > 0) & (An_GestDay <= An_GestLength):  # Line 2231-2232
        Fet_Wt = (Fet_BWbrth * 
                  np.exp(-(coeff_dict['Fet_Ksyn'] - 
                           coeff_dict['Fet_KsynDecay'] * An_GestDay) * 
                           (An_GestLength - An_GestDay))) # gestating animal
    else:
        Fet_Wt = 0  # open animal
    return Fet_Wt


def calculate_Fet_BWgain(
    An_GestDay: int, 
    An_GestLength: int, 
    Fet_Wt: float,
    coeff_dict: dict
) -> float:
    """
    Fet_BWgain: Fetal bodyweight gain (kg/d)
    """
    if (An_GestDay > 0) & (An_GestDay <= An_GestLength): # gestating animal
        Fet_BWgain = (coeff_dict['Fet_Ksyn'] - 
                      coeff_dict['Fet_KsynDecay'] * An_GestDay) * Fet_Wt 
    else:
        Fet_BWgain = 0  # open animal, kg/d
    return Fet_BWgain


def calculate_Gest_AA_g(
    Gest_NPuse_g: float, 
    Body_AA_TP: np.ndarray
) -> np.ndarray:
    """
    Gest_AA_g: AA deposited in gravid uterus (g/d)
    """
    Gest_AA_g = Gest_NPuse_g * Body_AA_TP / 100  # Line 2367-2376
    return Gest_AA_g


def calculate_Gest_EAA_g(Gest_AA_g: pd.Series) -> float:
    """
    Gest_EAA_g: EAA deposited in gravid uterus (g/d)
    """
    Gest_EAA_g = Gest_AA_g.sum()  # Line 2377-2378
    return Gest_EAA_g


def calculate_GestAA_AbsAA(
    Gest_AA_g: pd.Series,
    Abs_AA_g: pd.Series
) -> pd.Series:
    """
    GestAA_AbsAA: AA debosited in gravid uterus as a fraction of absorbed AA 
    """
    GestAA_AbsAA = Gest_AA_g / Abs_AA_g  # Line 2381-2390
    return GestAA_AbsAA



# read version from installed package
from importlib.metadata import version
__version__ = version("nasem_dairy")

# from nasem_dairy.dag.ModelDAG import ModelDAG
try:
    from nasem_dairy.dag.ModelDAG import ModelDAG
except ImportError: # pragma: no cover
    ModelDAG = None 

from nasem_dairy.model.utility import read_csv_input, read_json_input, demo, get_feed_data, select_feeds
from nasem_dairy.model_output.ModelOutput import ModelOutput
from nasem_dairy.model.nasem import nasem
from nasem_dairy.data.constants import coeff_dict, infusion_dict, MP_NP_efficiency_dict, mPrt_coeff_list, f_Imb
from nasem_dairy.nasem_equations.dry_matter_intake import (
    calculate_Kb_LateGest_DMIn,
    calculate_An_PrePartWklim,
    calculate_Dt_DMIn_Heif_LateGestInd,
    calculate_Dt_DMIn_Heif_LateGestPen,
    calculate_Dt_NDFdev_DMI,
    calculate_Dt_DMIn_Heif_NRCa,
    calculate_Dt_DMIn_Heif_NRCad,
    calculate_Dt_DMIn_Heif_H1,
    calculate_Dt_DMIn_Heif_H2,
    calculate_Dt_DMIn_Heif_HJ1,
    calculate_Dt_DMIn_Heif_HJ2,
    calculate_Dt_DMIn_Lact1,
    calculate_Dt_DMIn_Lact2,
    calculate_Dt_DMIn_BW_LateGest_i,
    calculate_Dt_DMIn_BW_LateGest_p,
    calculate_Dt_DMIn_DryCow1_FarOff,
    calculate_Dt_DMIn_DryCow1_Close,
    calculate_Dt_DMIn_DryCow2,
    calculate_Dt_DMIn_Calf1,
    calculate_Dt_DMIn_DryCow_AdjGest,
    calculate_Dt_DMIn
)

from nasem_dairy.nasem_equations.milk import (
    calculate_Trg_NEmilk_Milk,
    calculate_Mlk_NP_g,
    calculate_Mlk_CP_g,
    calculate_An_LactDay_MlkPred,
    calculate_Trg_Mlk_Fat,
    calculate_Trg_Mlk_Fat_g,
    calculate_Mlk_Fatemp_g,
    calculate_Mlk_Fat_g,
    calculate_Mlk_Fat,
    calculate_Mlk_NP,
    calculate_Mlk_Prod_comp,
    calculate_An_MPavail_Milk_Trg,
    calculate_Mlk_NP_MPalow_Trg_g,
    calculate_Mlk_Prod_MPalow,
    calculate_An_MEavail_Milk,
    calculate_Mlk_Prod_NEalow,
    calculate_MlkNP_Milk,
    calculate_Mlk_Prod,
    calculate_MlkFat_Milk,
    calculate_MlkNE_Milk,
    calculate_Mlk_NEout,
    calculate_Mlk_MEout,
    calculate_Mlk_NPmx,
    calculate_MlkNP_MlkNPmx,
    calculate_Mlk_CP,
    calculate_Mlk_AA_g,
    calculate_Mlk_EAA_g,
    calculate_MlkNP_AnMP,
    calculate_MlkAA_AbsAA,
    calculate_MlkEAA_AbsEAA,
    calculate_MlkNP_AnCP,
    calculate_MlkAA_DtAA,
    calculate_Mlk_MPUse_g,
    calculate_Trg_MilkLac,
    calculate_Trg_NEmilk_DEIn,
    calculate_Trg_MilkProd_EPcor,
    calculate_Mlk_Prod_NEalow_EPcor,
    calculate_Mlk_EPcorNEalow_DMIn,
    calculate_MlkNP_Milk_p,
    calculate_MlkFat_Milk_p,
    calculate_Mlk_NE_DE,
    calculate_MlkNP_Int,
    calculate_MlkNP_DEInp,
    calculate_MlkNP_NDF,
    calculate_MlkNP_AbsAA,
    calculate_MlkNP_AbsEAA,
    calculate_MlkNP_AbsNEAA,
    calculate_MlkNP_AbsOthAA,
    calculate_Trg_Mlk_NP,
    calculate_Mlk_AA_TP
)

from nasem_dairy.nasem_equations.nutrient_intakes import (
    calculate_TT_dcFdNDF_Lg,
    calculate_Fd_DNDF48,
    calculate_TT_dcFdNDF_48h,
    calculate_TT_dcFdNDF_Base,
    calculate_Fd_GE,
    calculate_Fd_DMIn,
    calculate_Fd_AFIn,
    calculate_Fd_For,
    calculate_Fd_ForWet,
    calculate_Fd_ForDry,
    calculate_Fd_Past,
    calculate_Fd_LiqClf,
    calculate_Fd_ForNDF,
    calculate_Fd_NDFnf,
    calculate_Fd_NPNCP,
    calculate_Fd_NPN,
    calculate_Fd_NPNDM,
    calculate_Fd_TP,
    calculate_Fd_fHydr_FA,
    calculate_Fd_FAhydr,
    calculate_Fd_NFC,
    calculate_Fd_rOM,
    calculate_Fd_GEIn,
    calculate_Fd_DigNDFIn_Base,
    calculate_Fd_NPNCPIn,
    calculate_Fd_NPNIn,
    calculate_Fd_NPNDMIn,
    calculate_Fd_CPAIn,
    calculate_Fd_CPBIn,
    calculate_Fd_CPBIn_For,
    calculate_Fd_CPBIn_Conc,
    calculate_Fd_CPCIn,
    calculate_Fd_CPIn_ClfLiq,
    calculate_Fd_CPIn_ClfDry,
    calculate_Fd_rdcRUPB,
    calculate_Fd_RUPBIn,
    calculate_Fd_RUPIn,
    calculate_Fd_RUP_CP,
    calculate_Fd_RUP,
    calculate_Fd_RDP,
    calculate_Fd_OMIn,
    calculate_Fd_DE_base_1,
    calculate_Fd_DE_base_2,
    calculate_Fd_DE_base,
    calculate_Fd_DEIn_base,
    calculate_Fd_DEIn_base_ClfLiq,
    calculate_Fd_DEIn_base_ClfDry,
    calculate_Fd_DMIn_ClfLiq,
    calculate_Fd_DE_ClfLiq,
    calculate_Fd_ME_ClfLiq,
    calculate_Fd_DMIn_ClfFor,
    calculate_Fd_PinorgIn,
    calculate_Fd_PorgIn,
    calculate_Fd_MgIn_min,
    calculate_Fd_acCa,
    calculate_Fd_acPtot,
    calculate_Fd_acMg,
    calculate_Fd_acNa,
    calculate_Fd_acK,
    calculate_Fd_acCl,
    calculate_Fd_absCaIn,
    calculate_Fd_absPIn,
    calculate_Fd_absMgIn_base,
    calculate_Fd_absNaIn,
    calculate_Fd_absKIn,
    calculate_Fd_absClIn,
    calculate_Fd_acCo,
    calculate_Fd_acCu,
    calculate_Fd_acFe,
    calculate_Fd_acMn,
    calculate_Fd_acZn,
    calculate_Fd_DigSt,
    calculate_Fd_DigStIn_Base,
    calculate_Fd_DigrOMt,
    calculate_Fd_idRUPIn,
    calculate_TT_dcFdFA,
    calculate_Fd_DigFAIn,
    calculate_Fd_DigrOMa,
    calculate_Fd_DigrOMaIn,
    calculate_Fd_DigWSC,
    calculate_Fd_DigWSCIn,
    calculate_Dt_ForDNDF48,
    calculate_Dt_ForDNDF48_ForNDF,
    calculate_Dt_ADF_NDF,
    calculate_Dt_DE_ClfLiq,
    calculate_Dt_ME_ClfLiq,
    calculate_Dt_NDFnfIn,
    calculate_Dt_Lg_NDF,
    calculate_Dt_ForNDFIn,
    calculate_Dt_PastSupplIn,
    calculate_Dt_NIn,
    calculate_Dt_RUPIn,
    calculate_Dt_RUP_CP,
    calculate_Dt_fCPBdu,
    calculate_Dt_UFAIn,
    calculate_Dt_MUFAIn,
    calculate_Dt_PUFAIn,
    calculate_Dt_SatFAIn,
    calculate_Dt_OMIn,
    calculate_Dt_rOMIn,
    calculate_Dt_DM,
    calculate_Dt_NDFIn_BW,
    calculate_Dt_ForNDF_NDF,
    calculate_Dt_ForNDFIn_BW,
    calculate_Dt_DMInSum,
    calculate_Dt_DEIn_ClfLiq,
    calculate_Dt_MEIn_ClfLiq,
    calculate_Dt_CPA_CP,
    calculate_Dt_CPB_CP,
    calculate_Dt_CPC_CP,
    calculate_Dt_DigNDFIn,
    calculate_Dt_DigStIn,
    calculate_Dt_DigrOMaIn,
    calculate_Dt_dcCP_ClfDry,
    calculate_Dt_DENDFIn,
    calculate_Dt_DEStIn,
    calculate_Dt_DErOMIn,
    calculate_Dt_DigCPaIn,
    calculate_Dt_DECPIn,
    calculate_Dt_DENPNCPIn,
    calculate_Dt_DETPIn,
    calculate_Dt_DEFAIn,
    calculate_Dt_DMIn_ClfStrt,
    calculate_Dt_DEIn,
    calculate_TT_dcNDF_Base,
    calculate_TT_dcNDF,
    calculate_TT_dcSt_Base,
    calculate_TT_dcSt,
    calculate_TT_dcAnSt,
    calculate_TT_dcrOMa,
    calculate_TT_dcrOMt,
    calculate_Dt_acMg,
    calculate_Abs_MgIn,
    calculate_Dt_DigWSCIn,
    calculate_Dt_DigSt,
    calculate_Dt_DigWSC,
    calculate_Dt_DigrOMa,
    calculate_Dt_DigrOMa_Dt,
    calculate_Dt_DigrOMt,
    calculate_Dt_DigNDFnfIn,
    calculate_Dt_DigNDF,
    calculate_Dt_DigNDFnf,
    calculate_Dt_DigCPtIn,
    calculate_Dt_DigTPaIn,
    calculate_Dt_DigTPtIn,
    calculate_Dt_DigCPa,
    calculate_TT_dcDtCPa,
    calculate_Dt_DigCPt,
    calculate_Dt_DigTPt,
    calculate_TT_dcDtCPt,
    calculate_Dt_MPIn,
    calculate_Dt_MP,
    calculate_Dt_DigUFAIn,
    calculate_Dt_DigMUFAIn,
    calculate_Dt_DigPUFAIn,
    calculate_Dt_DigSatFAIn,
    calculate_Dt_DigFA,
    calculate_Dt_DigOMaIn,
    calculate_Dt_DigOMtIn,
    calculate_Dt_DigOMa,
    calculate_Dt_DigOMt,
    calculate_TT_dcDtFA,
    calculate_Dt_GE,
    calculate_Dt_DE,
    calculate_Dt_TDN,
    calculate_Dt_TDNIn,
    calculate_Dt_GasE_IPCC2,
    calculate_Dt_GasEOut_Lact,
    calculate_Dt_GasEOut_Heif,
    calculate_Dt_GasEOut_Dry,
    calculate_Dt_GasEOut,
    calculate_feed_data,
    calculate_Fd_DigrOMtIn,
    calculate_Fd_idRUP,
    calculate_Fd_Fe_RUPout,
    calculate_Dt_RDPIn,
    calculate_Dt_idcRUP,
    calculate_Dt_Fe_RUPout,
    calculate_Dt_RDTPIn,
    calculate_Dt_RDP,
    calculate_Dt_RDP_CP,
    calculate_Fd_AFInp,
    calculate_Fd_RDPIn,
    calculate_diet_data,
    calculate_Dt_IdAARUPIn_array,
    calculate_Fd_ADFIn,
    calculate_Fd_NDFIn,
    calculate_Fd_StIn,
    calculate_Fd_NFCIn,
    calculate_Fd_WSCIn,
    calculate_Fd_rOMIn,
    calculate_Fd_LgIn,
    calculate_Fd_ConcIn,
    calculate_Fd_ForIn,
    calculate_Fd_ForNDFIn,
    calculate_Fd_ForWetIn,
    calculate_Fd_ForDryIn,
    calculate_Fd_PastIn,
    calculate_Fd_CPIn,
    calculate_Fd_TPIn,
    calculate_Fd_CFatIn,
    calculate_Fd_FAIn,
    calculate_Fd_FAhydrIn,
    calculate_Fd_AshIn,
    calculate_Fd_C120In,
    calculate_Fd_C140In,
    calculate_Fd_C160In,
    calculate_Fd_C161In,
    calculate_Fd_C180In,
    calculate_Fd_C181tIn,
    calculate_Fd_C181cIn,
    calculate_Fd_C182In,
    calculate_Fd_C183In,
    calculate_Fd_OtherFAIn,
    calculate_Fd_CaIn,
    calculate_Fd_PIn,
    calculate_Fd_NaIn,
    calculate_Fd_MgIn,
    calculate_Fd_KIn,
    calculate_Fd_ClIn,
    calculate_Fd_SIn,
    calculate_Fd_CoIn,
    calculate_Fd_CrIn,
    calculate_Fd_CuIn,
    calculate_Fd_FeIn,
    calculate_Fd_IIn,
    calculate_Fd_MnIn,
    calculate_Fd_MoIn,
    calculate_Fd_SeIn,
    calculate_Fd_ZnIn,
    calculate_Fd_VitAIn,
    calculate_Fd_VitDIn,
    calculate_Fd_VitEIn,
    calculate_Fd_CholineIn,
    calculate_Fd_BiotinIn,
    calculate_Fd_NiacinIn,
    calculate_Fd_B_CaroteneIn,
    calculate_Fd_absCoIn,
    calculate_Fd_absCuIn,
    calculate_Fd_absFeIn,
    calculate_Fd_absMnIn,
    calculate_Fd_absZnIn,
    calculate_Fd_Argt_CP,
    calculate_Fd_Hist_CP,
    calculate_Fd_Ilet_CP,
    calculate_Fd_Leut_CP,
    calculate_Fd_Lyst_CP,
    calculate_Fd_Mett_CP,
    calculate_Fd_Phet_CP,
    calculate_Fd_Thrt_CP,
    calculate_Fd_Trpt_CP,
    calculate_Fd_Valt_CP,
    calculate_Fd_ArgRUPIn,
    calculate_Fd_HisRUPIn,
    calculate_Fd_IleRUPIn,
    calculate_Fd_LeuRUPIn,
    calculate_Fd_LysRUPIn,
    calculate_Fd_MetRUPIn,
    calculate_Fd_PheRUPIn,
    calculate_Fd_ThrRUPIn,
    calculate_Fd_TrpRUPIn,
    calculate_Fd_ValRUPIn,
    calculate_Fd_IdArgRUPIn,
    calculate_Fd_IdHisRUPIn,
    calculate_Fd_IdIleRUPIn,
    calculate_Fd_IdLeuRUPIn,
    calculate_Fd_IdLysRUPIn,
    calculate_Fd_IdMetRUPIn,
    calculate_Fd_IdPheRUPIn,
    calculate_Fd_IdThrRUPIn,
    calculate_Fd_IdTrpRUPIn,
    calculate_Fd_IdValRUPIn,
    calculate_Fd_DigC120In,
    calculate_Fd_DigC140In,
    calculate_Fd_DigC160In,
    calculate_Fd_DigC161In,
    calculate_Fd_DigC180In,
    calculate_Fd_DigC181tIn,
    calculate_Fd_DigC181cIn,
    calculate_Fd_DigC182In,
    calculate_Fd_DigC183In,
    calculate_Fd_DigOtherFAIn,
    calculate_Fd_ArgIn,
    calculate_Fd_HisIn,
    calculate_Fd_IleIn,
    calculate_Fd_LeuIn,
    calculate_Fd_LysIn,
    calculate_Fd_MetIn,
    calculate_Fd_PheIn,
    calculate_Fd_ThrIn,
    calculate_Fd_TrpIn,
    calculate_Fd_ValIn,
    calculate_Dt_ADF,
    calculate_Dt_NDF,
    calculate_Dt_For,
    calculate_Dt_ForNDF,
    calculate_Dt_DMIn_ClfLiq,
    calculate_Dt_DMIn_ClfFor,
    calculate_Dt_AFIn,
    calculate_Dt_NDFIn,
    calculate_Dt_ADFIn,
    calculate_Dt_LgIn,
    calculate_Dt_DigNDFIn_Base,
    calculate_Dt_ForWetIn,
    calculate_Dt_ForDryIn,
    calculate_Dt_PastIn,
    calculate_Dt_ForIn,
    calculate_Dt_ConcIn,
    calculate_Dt_NFCIn,
    calculate_Dt_StIn,
    calculate_Dt_WSCIn,
    calculate_Dt_CPIn,
    calculate_Dt_CPIn_ClfLiq,
    calculate_Dt_TPIn,
    calculate_Dt_NPNCPIn,
    calculate_Dt_NPNIn,
    calculate_Dt_NPNDMIn,
    calculate_Dt_CPAIn,
    calculate_Dt_CPBIn,
    calculate_Dt_CPCIn,
    calculate_Dt_RUPBIn,
    calculate_Dt_CFatIn,
    calculate_Dt_FAIn,
    calculate_Dt_FAhydrIn,
    calculate_Dt_C120In,
    calculate_Dt_C140In,
    calculate_Dt_C160In,
    calculate_Dt_C161In,
    calculate_Dt_C180In,
    calculate_Dt_C181tIn,
    calculate_Dt_C181cIn,
    calculate_Dt_C182In,
    calculate_Dt_C183In,
    calculate_Dt_OtherFAIn,
    calculate_Dt_AshIn,
    calculate_Dt_GEIn,
    calculate_Dt_DEIn_base,
    calculate_Dt_DEIn_base_ClfLiq,
    calculate_Dt_DEIn_base_ClfDry,
    calculate_Dt_DigStIn_Base,
    calculate_Dt_DigrOMtIn,
    calculate_Dt_idRUPIn,
    calculate_Dt_DigFAIn,
    calculate_Dt_ArgIn,
    calculate_Dt_HisIn,
    calculate_Dt_IleIn,
    calculate_Dt_LeuIn,
    calculate_Dt_LysIn,
    calculate_Dt_MetIn,
    calculate_Dt_PheIn,
    calculate_Dt_ThrIn,
    calculate_Dt_TrpIn,
    calculate_Dt_ValIn,
    calculate_Dt_ArgRUPIn,
    calculate_Dt_HisRUPIn,
    calculate_Dt_IleRUPIn,
    calculate_Dt_LeuRUPIn,
    calculate_Dt_LysRUPIn,
    calculate_Dt_MetRUPIn,
    calculate_Dt_PheRUPIn,
    calculate_Dt_ThrRUPIn,
    calculate_Dt_TrpRUPIn,
    calculate_Dt_ValRUPIn,
    calculate_Dt_RUP,
    calculate_Dt_OM,
    calculate_Dt_NDFnf,
    calculate_Dt_Lg,
    calculate_Dt_NFC,
    calculate_Dt_St,
    calculate_Dt_WSC,
    calculate_Dt_rOM,
    calculate_Dt_CFat,
    calculate_Dt_FA,
    calculate_Dt_FAhydr,
    calculate_Dt_CP,
    calculate_Dt_TP,
    calculate_Dt_NPNCP,
    calculate_Dt_NPN,
    calculate_Dt_NPNDM,
    calculate_Dt_CPA,
    calculate_Dt_CPB,
    calculate_Dt_CPC,
    calculate_Dt_Ash,
    calculate_Dt_ForWet,
    calculate_Dt_ForDry,
    calculate_Dt_Conc,
    calculate_Dt_C120,
    calculate_Dt_C140,
    calculate_Dt_C160,
    calculate_Dt_C161,
    calculate_Dt_C180,
    calculate_Dt_C181t,
    calculate_Dt_C181c,
    calculate_Dt_C182,
    calculate_Dt_C183,
    calculate_Dt_OtherFA,
    calculate_Dt_UFA,
    calculate_Dt_MUFA,
    calculate_Dt_PUFA,
    calculate_Dt_SatFA,
    calculate_Dt_C120_FA,
    calculate_Dt_C140_FA,
    calculate_Dt_C160_FA,
    calculate_Dt_C161_FA,
    calculate_Dt_C180_FA,
    calculate_Dt_C181t_FA,
    calculate_Dt_C181c_FA,
    calculate_Dt_C182_FA,
    calculate_Dt_C183_FA,
    calculate_Dt_OtherFA_FA,
    calculate_Dt_UFA_FA,
    calculate_Dt_MUFA_FA,
    calculate_Dt_PUFA_FA,
    calculate_Dt_SatFA_FA,
    calculate_Dt_CaIn,
    calculate_Dt_PIn,
    calculate_Dt_PinorgIn,
    calculate_Dt_PorgIn,
    calculate_Dt_NaIn,
    calculate_Dt_MgIn,
    calculate_Dt_MgIn_min,
    calculate_Dt_KIn,
    calculate_Dt_ClIn,
    calculate_Dt_SIn,
    calculate_Dt_CoIn,
    calculate_Dt_CrIn,
    calculate_Dt_CuIn,
    calculate_Dt_FeIn,
    calculate_Dt_IIn,
    calculate_Dt_MnIn,
    calculate_Dt_MoIn,
    calculate_Dt_SeIn,
    calculate_Dt_ZnIn,
    calculate_Dt_VitAIn,
    calculate_Dt_VitDIn,
    calculate_Dt_VitEIn,
    calculate_Dt_CholineIn,
    calculate_Dt_BiotinIn,
    calculate_Dt_NiacinIn,
    calculate_Dt_B_CaroteneIn,
    calculate_Dt_Ca,
    calculate_Dt_P,
    calculate_Dt_Pinorg,
    calculate_Dt_Porg,
    calculate_Dt_Na,
    calculate_Dt_Mg,
    calculate_Dt_K,
    calculate_Dt_Cl,
    calculate_Dt_S,
    calculate_Dt_Co,
    calculate_Dt_Cr,
    calculate_Dt_Cu,
    calculate_Dt_Fe,
    calculate_Dt_I,
    calculate_Dt_Mn,
    calculate_Dt_Mo,
    calculate_Dt_Se,
    calculate_Dt_Zn,
    calculate_Dt_VitA,
    calculate_Dt_VitD,
    calculate_Dt_VitE,
    calculate_Dt_Choline,
    calculate_Dt_Biotin,
    calculate_Dt_Niacin,
    calculate_Dt_B_Carotene,
    calculate_Dt_IdArgRUPIn,
    calculate_Dt_IdHisRUPIn,
    calculate_Dt_IdIleRUPIn,
    calculate_Dt_IdLeuRUPIn,
    calculate_Dt_IdLysRUPIn,
    calculate_Dt_IdMetRUPIn,
    calculate_Dt_IdPheRUPIn,
    calculate_Dt_IdThrRUPIn,
    calculate_Dt_IdTrpRUPIn,
    calculate_Dt_IdValRUPIn,
    calculate_Abs_CaIn,
    calculate_Abs_PIn,
    calculate_Abs_NaIn,
    calculate_Abs_KIn,
    calculate_Abs_ClIn,
    calculate_Abs_CoIn,
    calculate_Abs_CuIn,
    calculate_Abs_FeIn,
    calculate_Abs_MnIn,
    calculate_Abs_ZnIn,
    calculate_Dt_DigFA_FA,
    calculate_Dt_DigC120_FA,
    calculate_Dt_DigC140_FA,
    calculate_Dt_DigC160_FA,
    calculate_Dt_DigC161_FA,
    calculate_Dt_DigC180_FA,
    calculate_Dt_DigC181t_FA,
    calculate_Dt_DigC181c_FA,
    calculate_Dt_DigC182_FA,
    calculate_Dt_DigC183_FA,
    calculate_Dt_DigUFA_FA,
    calculate_Dt_DigMUFA_FA,
    calculate_Dt_DigPUFA_FA,
    calculate_Dt_DigSatFA_FA,
    calculate_Dt_DigOtherFA_FA,
    calculate_Dt_IdArgIn,
    calculate_Dt_IdHisIn,
    calculate_Dt_IdIleIn,
    calculate_Dt_IdLeuIn,
    calculate_Dt_IdLysIn,
    calculate_Dt_IdMetIn,
    calculate_Dt_IdPheIn,
    calculate_Dt_IdThrIn,
    calculate_Dt_IdTrpIn,
    calculate_Dt_IdValIn,
    calculate_Fd_DMInp,
    calculate_Trg_Fd_DMIn,
    calculate_Dt_DigC120In,
    calculate_Dt_DigC140In,
    calculate_Dt_DigC160In,
    calculate_Dt_DigC161In,
    calculate_Dt_DigC180In,
    calculate_Dt_DigC181tIn,
    calculate_Dt_DigC181cIn,
    calculate_Dt_DigC182In,
    calculate_Dt_DigC183In,
    calculate_Dt_DigOtherFAIn,
    calculate_DtArgRUP_DtArg,
    calculate_DtHisRUP_DtHis,
    calculate_DtIleRUP_DtIle,
    calculate_DtLeuRUP_DtLeu,
    calculate_DtLysRUP_DtLys,
    calculate_DtMetRUP_DtMet,
    calculate_DtPheRUP_DtPhe,
    calculate_DtThrRUP_DtThr,
    calculate_DtTrpRUP_DtTrp,
    calculate_DtValRUP_DtVal
)

from nasem_dairy.nasem_equations.rumen import (
    calculate_Rum_dcNDF,
    calculate_Rum_dcSt,
    calculate_Rum_DigNDFIn,
    calculate_Rum_DigStIn,
    calculate_Rum_DigNDFnfIn,
    calculate_Du_StPas,
    calculate_Du_NDFPas
)

from nasem_dairy.nasem_equations.microbial_protein import (
    calculate_RDPIn_MiNmax,
    calculate_MiN_Vm,
    calculate_Du_MiN_NRC2021_g,
    calculate_Du_MiN_VTln_g,
    calculate_Du_MiN_VTnln_g,
    calculate_Du_MiCP,
    calculate_Du_idMiCP_g,
    calculate_Du_idMiCP,
    calculate_Du_idMiTP_g,
    calculate_Du_idMiTP,
    calculate_Du_MiTP,
    calculate_Du_EndCP_g,
    calculate_Du_EndN_g,
    calculate_Du_EndCP,
    calculate_Du_EndN,
    calculate_Du_NAN_g,
    calculate_Du_NANMN_g,
    calculate_Du_MiN_NRC2001_g,
    calculate_Rum_MiCP_DigCHO,
    calculate_Du_MiN_g
)

from nasem_dairy.nasem_equations.protein import (
    calculate_f_mPrt_max, 
    calculate_Du_MiCP_g, 
    calculate_Du_MiTP_g,
    calculate_Scrf_CP_g,
    calculate_Scrf_NP_g,
    calculate_Scrf_MPUse_g_Trg,
    calculate_Scrf_NP,
    calculate_Scrf_N_g,
    calculate_Scrf_AA_g,
    calculate_ScrfAA_AbsAA,
    calculate_An_CPxprt_g,
    calculate_An_NPxprt_g,
    calculate_Trg_NPxprt_g,
    calculate_An_CPprod_g,
    calculate_An_NPprod_g,
    calculate_Trg_NPprod_g,
    calculate_An_NPprod_MPIn,
    calculate_Trg_NPuse_g,
    calculate_An_NPuse_g,
    calculate_An_NCPuse_g,
    calculate_An_Nprod_g,
    calculate_An_Nprod_NIn,
    calculate_An_Nprod_DigNIn,
    calculate_An_MPBal_g_Trg,
    calculate_Xprt_NP_MP_Trg,
    calculate_Xprt_NP_MP,
    calculate_Km_MP_NP,
    calculate_Kl_MP_NP,
    calculate_Scrf_MPUse_g,
    calculate_An_MPuse_g,
    calculate_An_MPuse,
    calculate_An_MPBal_g,
    calculate_An_MP_NP,
    calculate_An_NPxprt_MP,
    calculate_An_CP_NP,
    calculate_An_NPBal_g,
    calculate_An_NPBal,
    calculate_Scrf_AA_TP
)

from nasem_dairy.nasem_equations.amino_acid import (
    calculate_Du_AAMic,
    calculate_Du_IdAAMic,
    calculate_Abs_AA_g,
    calculate_mPrtmx_AA,
    calculate_mPrtmx_AA2,
    calculate_AA_mPrtmx,
    calculate_mPrt_AA_01,
    calculate_mPrt_k_AA,
    calculate_Abs_EAA_g,
    calculate_Abs_neAA_g,
    calculate_Abs_OthAA_g,
    calculate_Abs_EAA2_HILKM_g,
    calculate_Abs_EAA2_RHILKM_g,
    calculate_Abs_EAA2_HILKMT_g,
    calculate_Abs_EAA2b_g,
    calculate_mPrt_k_EAA2,
    calculate_Du_AAEndP,
    calculate_Du_AA,
    calculate_DuAA_AArg,
    calculate_Du_AA24h,
    calculate_IdAA_DtAA,
    calculate_Abs_AA_MPp,
    calculate_Abs_AA_p,
    calculate_Abs_AA_DEI,
    calculate_Abs_AA_mol,
    calculate_Body_AAGain_g,
    calculate_Body_EAAGain_g,
    calculate_BodyAA_AbsAA,
    calculate_An_AAUse_g,
    calculate_An_EAAUse_g,
    calculate_AnAAUse_AbsAA,
    calculate_AnEAAUse_AbsEAA,
    calculate_An_AABal_g,
    calculate_An_EAABal_g,
    calculate_Trg_AbsEAA_NPxprtEAA,
    calculate_Trg_AbsArg_NPxprtArg,
    calculate_Trg_AAEff_EAAEff,
    calculate_An_AAEff_EAAEff,
    calculate_Imb_AA,
    calculate_Imb_EAA,
    calculate_An_IdEAAIn,
    calculate_Du_IdEAAMic,
    calculate_Dt_IdEAARUPIn,
    calculate_Trg_Mlk_AA_g,
    calculate_Trg_Mlk_EAA_g,
    calculate_Trg_AAUse_g,
    calculate_Trg_EAAUse_g,
    calculate_Trg_AbsAA_g,
    calculate_Trg_AbsEAA_g,
    calculate_Trg_MlkEAA_AbsEAA,
    calculate_AnNPxAA_AbsAA,
    calculate_AnNPxEAA_AbsEAA,
    calculate_AnNPxAAUser_AbsAA,
    calculate_AnNPxEAAUser_AbsEAA,
    calculate_MiTPAAProf,
    calculate_mPrt_k_AA_array,
    calculate_An_IdAAIn_array,
    calculate_Inf_AA_g,
    calculate_EndAAProf,
    calculate_Dt_AARUPIn,
    calculate_Inf_AARUPIn,
    calculate_Dt_AAIn,
    calculate_RecAA,
    calculate_MWAA,
    calculate_Body_AA_TP,
    calculate_Abs_EAA2_g,
    calculate_Trg_AbsAA_NPxprtAA_array,
    calculate_Du_EAA_g
)


from nasem_dairy.nasem_equations.infusion import (
    calculate_Inf_TPIn,
    calculate_Inf_OMIn,
    calculate_Inf_Rum,
    calculate_Inf_SI,
    calculate_Inf_Art,
    calculate_InfRum_TPIn,
    calculate_InfSI_TPIn,
    calculate_InfRum_RUPIn,
    calculate_InfRum_RUP_CP,
    calculate_InfRum_idRUPIn,
    calculate_InfSI_idTPIn,
    calculate_InfSI_idCPIn,
    calculate_Inf_idCPIn,
    calculate_InfRum_RDPIn,
    calculate_Inf_DigFAIn,
    calculate_Inf_DEAcetIn,
    calculate_Inf_DEPropIn,
    calculate_Inf_DEButrIn,
    calculate_infusion_data,
    calculate_Inf_DMIn,
    calculate_Inf_StIn,
    calculate_Inf_NDFIn,
    calculate_Inf_ADFIn,
    calculate_Inf_GlcIn,
    calculate_Inf_CPIn,
    calculate_Inf_NPNCPIn,
    calculate_Inf_FAIn,
    calculate_Inf_AshIn,
    calculate_Inf_VFAIn,
    calculate_Inf_AcetIn,
    calculate_Inf_PropIn,
    calculate_Inf_ButrIn,
    calculate_Inf_CPAIn,
    calculate_Inf_CPBIn,
    calculate_Inf_CPCIn,
    calculate_Inf_DM,
    calculate_Inf_OM,
    calculate_Inf_St,
    calculate_Inf_NDF,
    calculate_Inf_ADF,
    calculate_Inf_Glc,
    calculate_Inf_CP,
    calculate_Inf_FA,
    calculate_Inf_VFA,
    calculate_Inf_Acet,
    calculate_Inf_Prop,
    calculate_Inf_Butr,
    calculate_InfRum_DMIn,
    calculate_InfRum_OMIn,
    calculate_InfRum_CPIn,
    calculate_InfRum_NPNCPIn,
    calculate_InfRum_CPAIn,
    calculate_InfRum_CPBIn,
    calculate_InfRum_CPCIn,
    calculate_InfRum_StIn,
    calculate_InfRum_NDFIn,
    calculate_InfRum_ADFIn,
    calculate_InfRum_FAIn,
    calculate_InfRum_GlcIn,
    calculate_InfRum_VFAIn,
    calculate_InfRum_AcetIn,
    calculate_InfRum_PropIn,
    calculate_InfRum_ButrIn,
    calculate_InfRum_AshIn,
    calculate_InfSI_DMIn,
    calculate_InfSI_OMIn,
    calculate_InfSI_CPIn,
    calculate_InfSI_NPNCPIn,
    calculate_InfSI_StIn,
    calculate_InfSI_GlcIn,
    calculate_InfSI_NDFIn,
    calculate_InfSI_ADFIn,
    calculate_InfSI_FAIn,
    calculate_InfSI_VFAIn,
    calculate_InfSI_AcetIn,
    calculate_InfSI_PropIn,
    calculate_InfSI_ButrIn,
    calculate_InfSI_AshIn,
    calculate_InfArt_DMIn,
    calculate_InfArt_OMIn,
    calculate_InfArt_CPIn,
    calculate_InfArt_NPNCPIn,
    calculate_InfArt_TPIn,
    calculate_InfArt_StIn,
    calculate_InfArt_GlcIn,
    calculate_InfArt_NDFIn,
    calculate_InfArt_ADFIn,
    calculate_InfArt_FAIn,
    calculate_InfArt_VFAIn,
    calculate_InfArt_AcetIn,
    calculate_InfArt_PropIn,
    calculate_InfArt_ButrIn,
    calculate_InfArt_AshIn,
    calculate_Inf_ArgRUPIn,
    calculate_Inf_HisRUPIn,
    calculate_Inf_IleRUPIn,
    calculate_Inf_LeuRUPIn,
    calculate_Inf_LysRUPIn,
    calculate_Inf_MetRUPIn,
    calculate_Inf_PheRUPIn,
    calculate_Inf_ThrRUPIn,
    calculate_Inf_TrpRUPIn,
    calculate_Inf_ValRUPIn,
    calculate_Inf_IdArgIn,
    calculate_Inf_IdHisIn,
    calculate_Inf_IdIleIn,
    calculate_Inf_IdLeuIn,
    calculate_Inf_IdLysIn,
    calculate_Inf_IdMetIn,
    calculate_Inf_IdPheIn,
    calculate_Inf_IdThrIn,
    calculate_Inf_IdTrpIn,
    calculate_Inf_IdValIn,
)

from nasem_dairy.nasem_equations.animal import (
    calculate_An_DMIn_BW,
    calculate_An_RDPIn,
    calculate_An_RDP,
    calculate_An_RDPIn_g,
    calculate_An_NDFIn,
    calculate_An_NDF,
    calculate_An_DigNDFIn,
    calculate_An_DENDFIn,
    calculate_An_DigStIn,
    calculate_An_DEStIn,
    calculate_An_DigrOMaIn,
    calculate_An_DErOMIn,
    calculate_An_idRUPIn,
    calculate_An_RUPIn,
    calculate_An_DMIn,
    calculate_An_CPIn,
    calculate_An_DigNDF,
    calculate_An_GasEOut_Dry,
    calculate_An_GasEOut_Lact,
    calculate_An_GasEOut_Heif,
    calculate_An_GasEOut,
    calculate_An_DigCPaIn,
    calculate_An_DECPIn,
    calculate_An_DENPNCPIn,
    calculate_An_DETPIn,
    calculate_An_DigFAIn,
    calculate_An_DEFAIn,
    calculate_An_DEIn,
    calculate_An_DEInp,
    calculate_An_GutFill_BW,
    calculate_An_BWnp,
    calculate_An_GutFill_Wt,
    calculate_An_BW_empty,
    calculate_An_REgain_Calf,
    calculate_An_MEIn_approx,
    calculate_An_MEIn,
    calculate_An_NEIn,
    calculate_An_NE,
    calculate_An_MBW,
    calculate_An_TPIn,
    calculate_An_DigTPaIn,
    calculate_An_MPIn,
    calculate_An_MPIn_g,
    calculate_An_DigStIn_Base,
    calculate_An_DigWSCIn,
    calculate_An_DigrOMtIn,
    calculate_An_DigSt,
    calculate_An_DigWSC,
    calculate_An_DigrOMa,
    calculate_An_DigrOMt,
    calculate_An_DigNDFIn_Base,
    calculate_An_RDPbal_g,
    calculate_An_DigCPa,
    calculate_TT_dcAnCPa,
    calculate_An_DigCPtIn,
    calculate_An_DigNtIn_g,
    calculate_An_DigTPtIn,
    calculate_An_DigCPt,
    calculate_An_DigTPt,
    calculate_TT_dcAnCPt,
    calculate_TT_dcAnTPt,
    calculate_SI_dcAnRUP,
    calculate_An_idCPIn,
    calculate_An_MP,
    calculate_An_MP_CP,
    calculate_An_DigFA,
    calculate_TT_dcAnFA,
    calculate_An_OMIn,
    calculate_An_DigOMaIn_Base,
    calculate_An_DigOMtIn_Base,
    calculate_An_DigOMaIn,
    calculate_An_DigOMtIn,
    calculate_TT_dcOMa,
    calculate_TT_dcOMt,
    calculate_TT_dcOMt_Base,
    calculate_An_DigOMa,
    calculate_An_DigOMt,
    calculate_An_GEIn,
    calculate_An_GE,
    calculate_An_DERDTPIn,
    calculate_An_DEidRUPIn,
    calculate_An_DE,
    calculate_An_DE_GE,
    calculate_An_DEnp,
    calculate_An_GasE_IPCC2,
    calculate_GasE_DMIn,
    calculate_GasE_GEIn,
    calculate_GasE_DEIn,
    calculate_An_NPm_Use,
    calculate_An_CPm_Use,
    calculate_An_ME,
    calculate_An_ME_GE,
    calculate_An_ME_DE,
    calculate_An_NE_GE,
    calculate_An_NE_DE,
    calculate_An_NE_ME,
    calculate_An_MPIn_MEIn,
    calculate_An_RUPIn_g,
    calculate_An_MEIn_ClfDry,
    calculate_An_ME_ClfDry,
    calculate_An_NE_ClfDry,
    calculate_An_Grazing,
    calculate_En_OM,
    calculate_An_DMIn_MBW,
    calculate_An_StIn,
    calculate_An_St,
    calculate_An_rOMIn,
    calculate_An_rOM,
    calculate_An_NDFIn_BW,
    calculate_An_ADFIn,
    calculate_An_ADF,
    calculate_An_CPIn_g,
    calculate_An_CP,
    calculate_An_NIn_g,
    calculate_An_FAhydrIn,
    calculate_An_FA,
    calculate_An_AshIn,
    calculate_An_Ash,
    calculate_an_data,
    calculate_An_PrePartDay,
    calculate_An_PrePartWk,
    calculate_An_PrePartWkDurat,
    calculate_An_IdArgIn,
    calculate_An_IdHisIn,
    calculate_An_IdIleIn,
    calculate_An_IdLeuIn,
    calculate_An_IdLysIn,
    calculate_An_IdMetIn,
    calculate_An_IdPheIn,
    calculate_An_IdThrIn,
    calculate_An_IdTrpIn,
    calculate_An_IdValIn,
    calculate_An_NPNCPIn,
    calculate_An_FAIn,
    calculate_An_RDNPNCPIn,
    calculate_An_RUP,
    calculate_An_RUP_CP,
    calculate_An_idRUCPIn,
    calculate_An_idRUP,
    calculate_An_RDTPIn,
    calculate_An_RDP_CP
)

from nasem_dairy.nasem_equations.gestation import (
    calculate_Uter_Wtpart,
    calculate_Uter_Wt,
    calculate_GrUter_Wtpart,
    calculate_GrUter_Wt,
    calculate_Uter_BWgain,
    calculate_GrUter_BWgain,
    calculate_Gest_NCPgain_g,
    calculate_Gest_NPgain_g,
    calculate_Gest_NPuse_g,
    calculate_Gest_CPuse_g,
    calculate_An_PostPartDay,
    calculate_An_Preg,
    calculate_Fet_Wt,
    calculate_Fet_BWgain,
    calculate_Gest_AA_g,
    calculate_Gest_EAA_g,
    calculate_GestAA_AbsAA
)

from nasem_dairy.nasem_equations.fecal import (
    calculate_Fe_rOMend,
    calculate_Fe_RUP,
    calculate_Fe_RumMiCP,
    calculate_Fe_CPend_g,
    calculate_Fe_CPend,
    calculate_Fe_CP,
    calculate_Fe_NPend,
    calculate_Fe_NPend_g,
    calculate_Fe_MPendUse_g_Trg,
    calculate_Fe_rOM,
    calculate_Fe_St,
    calculate_Fe_NDF,
    calculate_Fe_NDFnf,
    calculate_Fe_Nend,
    calculate_Fe_RDPend,
    calculate_Fe_RUPend,
    calculate_Fe_MiTP,
    calculate_Fe_InfCP,
    calculate_Fe_TP,
    calculate_Fe_N,
    calculate_Fe_N_g,
    calculate_Fe_FA,
    calculate_Fe_OM,
    calculate_Fe_OM_end,
    calculate_Fe_DEMiCPend,
    calculate_Fe_DERDPend,
    calculate_Fe_DERUPend,
    calculate_Fe_DEout,
    calculate_Fe_DE_GE,
    calculate_Fe_DE,
    calculate_Fe_AAMet_g,
    calculate_Fe_AAMet_AbsAA,
    calculate_Fe_MPendUse_g,
    calculate_K_FeCPend_ClfLiq,
    calculate_Fe_AAMetab_TP
)   

from nasem_dairy.nasem_equations.body_composition import (
    calculate_CPGain_FrmGain,
    calculate_Frm_Gain,
    calculate_Frm_Gain_empty,
    calculate_NPGain_FrmGain,
    calculate_Rsrv_Gain,
    calculate_Rsrv_Gain_empty,
    calculate_Body_Gain_empty,
    calculate_Frm_NPgain,
    calculate_NPGain_RsrvGain,
    calculate_Rsrv_NPgain,
    calculate_Body_NPgain,
    calculate_Body_CPgain,
    calculate_Body_CPgain_g,
    calculate_Rsrv_Gain,
    calculate_Rsrv_Gain_empty,
    calculate_Rsrv_Fatgain,
    calculate_CPGain_FrmGain,
    calculate_Rsrv_CPgain,
    calculate_FatGain_FrmGain,
    calculate_Frm_Gain,
    calculate_Frm_Gain_empty,
    calculate_Frm_Fatgain,
    calculate_NPGain_FrmGain,
    calculate_Frm_NPgain,
    calculate_Frm_CPgain,
    calculate_Body_NPgain_g,
    calculate_An_BWmature_empty,
    calculate_Body_Gain,
    calculate_Trg_BWgain,
    calculate_Trg_BWgain_g,
    calculate_Conc_BWgain,
    calculate_BW_BCS,
    calculate_An_BWnp3,
    calculate_An_GutFill_Wt_Erdman,
    calculate_An_BWnp_empty,
    calculate_An_BWnp3_empty,
    calculate_Body_Fat_EBW,
    calculate_Body_NonFat_EBW,
    calculate_Body_CP_EBW,
    calculate_Body_Ash_EBW,
    calculate_Body_Wat_EBW,
    calculate_Body_Fat,
    calculate_Body_NonFat,
    calculate_Body_CP,
    calculate_Body_Ash,
    calculate_Body_Wat,
    calculate_An_BodConcgain,
    calculate_NonFatGain_FrmGain,
    calculate_Body_Fatgain,
    calculate_Body_NonFatGain,
    calculate_Frm_CPgain_g,
    calculate_Rsrv_CPgain_g,
    calculate_Body_AshGain,
    calculate_Frm_AshGain,
    calculate_WatGain_RsrvGain,
    calculate_Rsrv_WatGain,
    calculate_Body_WatGain,
    calculate_Frm_WatGain,
    calculate_An_MPavail_Gain_Trg,
    calculate_Body_NPgain_MPalowTrg_g,
    calculate_Body_CPgain_MPalowTrg_g,
    calculate_Body_Gain_MPalowTrg_g,
    calculate_Body_Gain_MPalowTrg,
    calculate_An_MEavail_Grw,
    calculate_Kg_ME_NE,
    calculate_Body_Gain_NEalow,
    calculate_An_BodConcgain_NEalow,
    calculate_Body_Fatgain_NEalow,
    calculate_Body_NPgain_NEalow,
    calculate_An_Days_BCSdelta1,
    calculate_Rsrv_AshGain
)

from nasem_dairy.nasem_equations.urine import (
    calculate_Ur_Nout_g,
    calculate_Ur_DEout,
    calculate_Ur_Nend_g,
    calculate_Ur_NPend_g,
    calculate_Ur_MPendUse_g,
    calculate_Ur_Nend_Urea_g,
    calculate_Ur_Nend_Creatn_g,
    calculate_Ur_Nend_Creat_g,
    calculate_Ur_Nend_PD_g,
    calculate_Ur_NPend_3MH_g,
    calculate_Ur_Nend_3MH_g,
    calculate_Ur_Nend_sum_g,
    calculate_Ur_Nend_Hipp_g,
    calculate_Ur_NPend,
    calculate_Ur_MPend,
    calculate_Ur_EAAend_g,
    calculate_Ur_AAEnd_g,
    calculate_Ur_AAEnd_AbsAA,
    calculate_Ur_EAAEnd_g,
    calculate_Ur_Nout_DigNIn,
    calculate_Ur_Nout_CPcatab,
    calculate_UrDE_DMIn,
    calculate_UrDE_GEIn,
    calculate_UrDE_DEIn,
    calculate_Ur_AAEnd_TP
)

from nasem_dairy.nasem_equations.energy_requirement import (
    calculate_An_NEmUse_NS,
    calculate_An_NEm_Act_Graze,
    calculate_An_NEm_Act_Parlor,
    calculate_An_NEm_Act_Topo,
    calculate_An_NEmUse_Act,
    calculate_An_NEmUse,
    calculate_An_MEmUse,
    calculate_Rsrv_NEgain,
    calculate_Kr_ME_RE,
    calculate_Rsrv_MEgain,
    calculate_Frm_NEgain,
    calculate_Frm_MEgain,
    calculate_An_MEgain,
    calculate_Gest_REgain,
    calculate_Gest_MEuse,
    calculate_Trg_Mlk_NEout,
    calculate_Trg_Mlk_MEout,
    calculate_Trg_MEuse,
    calculate_An_MEmUse_NS,
    calculate_An_MEmUse_Act,
    calculate_An_MEmUse_Env,
    calculate_An_NEm_ME,
    calculate_An_NEm_DE,
    calculate_An_NEmNS_DE,
    calculate_An_NEmAct_DE,
    calculate_An_NEmEnv_DE,
    calculate_An_NEprod_Avail,
    calculate_An_MEprod_Avail,
    calculate_Gest_NELuse,
    calculate_Gest_NE_ME,
    calculate_Gest_NE_DE,
    calculate_An_REgain,
    calculate_Rsrv_NE_DE,
    calculate_Frm_NE_DE,
    calculate_Body_NEgain_BWgain,
    calculate_An_ME_NEg,
    calculate_Rsrv_NELgain,
    calculate_Frm_NELgain,
    calculate_An_NELgain,
    calculate_An_NEgain_DE,
    calculate_An_NEgain_ME,
    calculate_An_MEuse,
    calculate_An_NEuse,
    calculate_Trg_NEuse,
    calculate_An_NELuse,
    calculate_Trg_NELuse,
    calculate_An_NEprod_GE,
    calculate_Trg_NEprod_GE,
    calculate_An_NEmlk_GE,
    calculate_Trg_NEmlk_GE,
    calculate_An_MEbal,
    calculate_An_NELbal,
    calculate_An_NEbal,
    calculate_Trg_MEbal,
    calculate_Trg_NELbal,
    calculate_Trg_NEbal,
    calculate_An_MPuse_MEuse,
    calculate_Trg_MPuse_MEuse,
    calculate_Km_ME_NE_Clf,
    calculate_Km_ME_NE,
    calculate_Trg_NEmilkOut,
    calculate_Ky_ME_NE,
    calculate_Kf_ME_RE_ClfDry,
    calculate_Kf_ME_RE
)

from nasem_dairy.nasem_equations.protein_requirement import (
    calculate_An_MPm_g_Trg,
    calculate_Body_MPUse_g_Trg_initial,
    calculate_Gest_MPUse_g_Trg,
    calculate_Trg_Mlk_NP_g,
    calculate_Mlk_MPUse_g_Trg,
    calculate_An_MPuse_g_Trg_initial,
    calculate_Min_MPuse_g,
    calculate_Diff_MPuse_g,
    calculate_Frm_MPUse_g_Trg,
    calculate_Frm_NPgain_g,
    calculate_Kg_MP_NP_Trg_initial,
    calculate_Rsrv_NPgain_g,
    calculate_Rsrv_MPUse_g_Trg,
    calculate_Body_MPUse_g_Trg,
    calculate_An_MPuse_g_Trg,
    calculate_Trg_MPIn_req,
    calculate_Km_MP_NP_Trg,
    calculate_Kg_MP_NP_Trg_heifer_adjustment
)

from nasem_dairy.nasem_equations.micronutrient_requirement import (
    calculate_Ca_Mlk,
    calculate_Fe_Ca_m,
    calculate_An_Ca_g,
    calculate_An_Ca_y,
    calculate_An_Ca_l,
    calculate_An_Ca_Clf,
    calculate_An_Ca_req,
    calculate_An_Ca_bal,
    calculate_An_Ca_prod,
    calculate_Ur_P_m,
    calculate_Fe_P_m,
    calculate_An_P_m,
    calculate_An_P_g,
    calculate_An_P_y,
    calculate_An_P_l,
    calculate_An_P_Clf,
    calculate_An_P_req,
    calculate_An_P_bal,
    calculate_Fe_P_g,
    calculate_An_P_prod,
    calculate_An_Mg_Clf,
    calculate_Ur_Mg_m,
    calculate_Fe_Mg_m,
    calculate_An_Mg_m,
    calculate_An_Mg_g,
    calculate_An_Mg_y,
    calculate_An_Mg_l,
    calculate_An_Mg_req,
    calculate_An_Mg_bal,
    calculate_An_Mg_prod,
    calculate_An_Na_Clf,
    calculate_Fe_Na_m,
    calculate_An_Na_g,
    calculate_An_Na_y,
    calculate_An_Na_l,
    calculate_An_Na_req,
    calculate_An_Na_bal,
    calculate_An_Na_prod,
    calculate_An_Cl_Clf,
    calculate_Fe_Cl_m,
    calculate_An_Cl_g,
    calculate_An_Cl_y,
    calculate_An_Cl_l,
    calculate_An_Cl_req,
    calculate_An_Cl_bal,
    calculate_An_Cl_prod,
    calculate_An_K_Clf,
    calculate_Ur_K_m,
    calculate_Fe_K_m,
    calculate_An_K_m,
    calculate_An_K_g,
    calculate_An_K_y,
    calculate_An_K_l,
    calculate_An_K_req,
    calculate_An_K_bal,
    calculate_An_K_prod,
    calculate_An_S_req,
    calculate_An_S_bal,
    calculate_An_Co_req,
    calculate_An_Co_bal,
    calculate_An_Cu_Clf,
    calculate_An_Cu_m,
    calculate_An_Cu_g,
    calculate_An_Cu_y,
    calculate_An_Cu_l,
    calculate_An_Cu_req,
    calculate_An_Cu_bal,
    calculate_An_Cu_prod,
    calculate_An_I_req,
    calculate_An_I_bal,
    calculate_An_Fe_Clf,
    calculate_An_Fe_g,
    calculate_An_Fe_y,
    calculate_An_Fe_l,
    calculate_An_Fe_req,
    calculate_An_Fe_bal,
    calculate_An_Fe_prod,
    calculate_An_Mn_Clf,
    calculate_An_Mn_m,
    calculate_An_Mn_g,
    calculate_An_Mn_y,
    calculate_An_Mn_l,
    calculate_An_Mn_req,
    calculate_An_Mn_bal,
    calculate_An_Mn_prod,
    calculate_An_Se_req,
    calculate_An_Se_bal,
    calculate_An_Zn_Clf,
    calculate_An_Zn_m,
    calculate_An_Zn_g,
    calculate_An_Zn_y,
    calculate_An_Zn_l,
    calculate_An_Zn_req,
    calculate_An_Zn_bal,
    calculate_An_Zn_prod,
    calculate_An_DCADmeq,
    calculate_An_VitA_req,
    calculate_An_VitA_bal,
    calculate_An_VitD_req,
    calculate_An_VitD_bal,
    calculate_An_VitE_req,
    calculate_An_VitE_bal,
    calculate_Dt_acCa,
    calculate_Dt_acP,
    calculate_Dt_acNa,
    calculate_Dt_acMg_final,
    calculate_Dt_acK,
    calculate_Dt_acCl,
    calculate_Dt_acCo,
    calculate_Dt_acCu,
    calculate_Dt_acFe,
    calculate_Dt_acMn,
    calculate_Dt_acZn,
    calculate_CaProd_CaIn,
    calculate_PProd_PIn,
    calculate_MgProd_MgIn,
    calculate_KProd_KIn,
    calculate_NaProd_NaIn,
    calculate_ClProd_ClIn,
    calculate_CuProd_CuIn,
    calculate_FeProd_FeIn,
    calculate_MnProd_MnIn,
    calculate_ZnProd_ZnIn,
    calculate_CaProd_CaAbs,
    calculate_PProd_PAbs,
    calculate_MgProd_MgAbs,
    calculate_KProd_KAbs,
    calculate_NaProd_NaAbs,
    calculate_ClProd_ClAbs,
    calculate_CuProd_CuAbs,
    calculate_FeProd_FeAbs,
    calculate_MnProd_MnAbs,
    calculate_ZnProd_ZnAbs,
    calculate_Dt_CaReq_DMI,
    calculate_Dt_PReq_DMI,
    calculate_Dt_MgReq_DMI,
    calculate_Dt_KReq_DMI,
    calculate_Dt_NaReq_DMI,
    calculate_Dt_ClReq_DMI,
    calculate_Dt_SReq_DMI,
    calculate_Dt_CoReq_DMI,
    calculate_Dt_CuReq_DMI,
    calculate_Dt_FeReq_DMI,
    calculate_Dt_IReq_DMI,
    calculate_Dt_MnReq_DMI,
    calculate_Dt_SeReq_DMI,
    calculate_Dt_ZnReq_DMI,
    calculate_Dt_VitAReq_DMI,
    calculate_Dt_VitDReq_DMI,
    calculate_Dt_VitEReq_DMI
)

from nasem_dairy.nasem_equations.coefficient_adjustment import adjust_LCT
from nasem_dairy.nasem_equations.report import (
    calculate_Dt_DMIn_BW,
    calculate_Dt_DMIn_MBW,
    calculate_percent_first_parity,
    calculate_age_first_calving,
    calculate_milk_lactose_percent,
    calculate_dmi_percent_bodyweight,
    calculate_adf_per_ndf,
    calculate_digestable_rup,
    calculate_Fd_AFIn_sum,
    calculate_Fd_DMIn_sum,
    calculate_Fe_DE_GE_percent,
    calculate_An_DE_GE_percent,
    calculate_UrDE_GEIn_percent,
    calculate_GasE_GEIn_percent,
    calculate_An_ME_GE_percent,
    calculate_An_NE_GE_percent,
    calculate_UrDE_DEIn_percent,
    calculate_GasE_DEIn_percent,
    calculate_An_ME_DE_percent,
    calculate_An_NE_DE_percent,
    calculate_An_NE_ME_percent,
    calculate_An_DEIn_percent,
    calculate_An_MEIn_percent,
    calculate_Dt_C120In_g,
    calculate_Dt_C140In_g,
    calculate_Dt_C160In_g,
    calculate_Dt_C161In_g,
    calculate_Dt_C180In_g,
    calculate_Dt_C181tIn_g,
    calculate_Dt_C181cIn_g,
    calculate_Dt_C182In_g,
    calculate_Dt_C183In_g,
    calculate_Dt_OtherFAIn_g,
    calculate_Dt_FAIn_g,
    calculate_Dt_SatFAIn_g,
    calculate_Dt_UFAIn_g,
    calculate_Dt_MUFAIn_g,
    calculate_Dt_PUFAIn_g,
    calculate_Dt_DigFAIn_g,
    calculate_An_RDPbal_kg,
    calculate_MP_from_body,
    calculate_An_BW_centered,
    calculate_An_DigNDF_centered,
    calculate_An_BW_protein,
    calculate_Dt_acCa_per_100g,
    calculate_Dt_acP_per_100g,
    calculate_Dt_acMg_per_100g,
    calculate_Dt_acCl_per_100g,
    calculate_Dt_acK_per_100g,
    calculate_Dt_acNa_per_100g,
    calculate_Dt_acCo_per_100g,
    calculate_Dt_acCu_per_100g,
    calculate_Dt_acFe_per_100g,
    calculate_Dt_acMn_per_100g,
    calculate_Dt_acZn_per_100g,
    calculate_An_MPuse_kg_Trg,
    calculate_Dt_ForNDFIn_percNDF
)

from nasem_dairy.nasem_equations.water import (
    calculate_An_WaIn,
    calculate_An_Wa_Insens,
    calculate_WaIn_Milk
)

from nasem_dairy.nasem_equations.methane import (
    calculate_CH4out_g,
    calculate_CH4out_L,
    calculate_CH4g_Milk,
    calculate_CH4L_Milk
)

from nasem_dairy.nasem_equations.manure import (
    calculate_Man_out,
    calculate_Man_Milk,
    calculate_Man_VolSld,
    calculate_Man_VolSld2,
    calculate_VolSlds_Milk,
    calculate_VolSlds_Milk2,
    calculate_Man_Nout_g,
    calculate_Man_Nout2_g,
    calculate_ManN_Milk,
    calculate_Man_Ca_out,
    calculate_Man_P_out,
    calculate_Man_Mg_out,
    calculate_Man_K_out,
    calculate_Man_Na_out,
    calculate_Man_Cl_out,
    calculate_Man_MacMin_out,
    calculate_Man_Cu_out,
    calculate_Man_Fe_out,
    calculate_Man_Mn_out,
    calculate_Man_Zn_out,
    calculate_Man_MicMin_out,
    calculate_Man_Min_out_g,
    calculate_Man_Wa_out,
    calculate_ManWa_Milk,
    calculate_VolSlds2_Milk
)


"""Functions for adjusting coefficient values used in the model. 
"""

def adjust_LCT(An_AgeDay: int) -> int:
    """
    An_AgeDay: Animal age in days
    LCT: Lower critical temperature in degrees celcius
    """
    if An_AgeDay > 21:
        LCT = 5  # calf > 3 wks of age, Line 229
    else:
        LCT = 15
    return LCT



"""Microbial protein synthesis and flow calculations.

This module contains functions to estimate the production, flow, and 
digestion of microbial protein within the rumen, based on various dietary inputs.
"""


def calculate_RDPIn_MiNmax(
    Dt_DMIn: float, 
    An_RDP: float, 
    An_RDPIn: float
) -> float:
    if An_RDP <= 12:  # Line 1124
        RDPIn_MiNmax = An_RDPIn
    else:
        RDPIn_MiNmax = Dt_DMIn * 0.12
    return RDPIn_MiNmax


def calculate_MiN_Vm(RDPIn_MiNmax: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"VmMiNInt": 100.8, "VmMiNRDPSlp": 81.56}
    
    calculate_MiN_Vm(
        RDPIn_MiNmax = 100.0, coeff_dict = coeff_dict
    )
    ```
    """
    MiN_Vm = coeff_dict['VmMiNInt'] + coeff_dict['VmMiNRDPSlp'] * RDPIn_MiNmax     
    # Line 1125
    return MiN_Vm


def calculate_Du_MiN_NRC2021_g(
    MiN_Vm: float, 
    Rum_DigNDFIn: float, 
    Rum_DigStIn: float, 
    An_RDPIn_g: float,
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"KmMiNRDNDF": 0.0939, "KmMiNRDSt": 0.0274}
    
    calculate_Du_MiN_NRC2021_g(
        MiN_Vm = 25.0, Rum_DigNDFIn = 200.0, Rum_DigStIn = 150.0, 
        An_RDPIn_g = 300.0, coeff_dict = coeff_dict
    )
    ```
    """
    Du_MiN_NRC2021_g = (MiN_Vm / (1 + coeff_dict['KmMiNRDNDF'] / Rum_DigNDFIn + 
                                  coeff_dict['KmMiNRDSt'] / Rum_DigStIn)) # Line 1126
    if Du_MiN_NRC2021_g > 1 * An_RDPIn_g / 6.25:  # Line 1130
        Du_MiN_NRC2021_g = 1 * An_RDPIn_g / 6.25
    else:
        Du_MiN_NRC2021_g = Du_MiN_NRC2021_g
    return Du_MiN_NRC2021_g


def calculate_Du_MiN_VTln_g(
    Dt_rOMIn: float, 
    Dt_ForNDFIn: float, 
    An_RDPIn: float, 
    Rum_DigStIn: float,
    Rum_DigNDFIn: float, 
    coeff_dict: dict
) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {
        "Int_MiN_VT": 18.686, "KrdSt_MiN_VT": 10.214, "KrdNDF_MiN_VT": 28.976, 
        "KRDP_MiN_VT": 43.405, "KrOM_MiN_VT": -11.731, "KForNDF_MiN_VT": 8.895, 
        "KrOM2_MiN_VT": 2.861, "KrdStxrOM_MiN_VT": 5.637, 
        "KrdNDFxForNDF_MiN_VT": -2.22
    }
    
    calculate_Du_MiN_VTln_g(
        Dt_rOMIn = 300.0, Dt_ForNDFIn = 100.0, An_RDPIn = 200.0, 
        Rum_DigStIn = 150.0, Rum_DigNDFIn = 250.0, coeff_dict = coeff_dict
    )
    ```
    """
    # Line 1144-1146
    Du_MiN_VTln_g = (coeff_dict['Int_MiN_VT'] + 
                     coeff_dict['KrdSt_MiN_VT'] * Rum_DigStIn + 
                     coeff_dict['KrdNDF_MiN_VT'] * Rum_DigNDFIn + 
                     coeff_dict['KRDP_MiN_VT'] * An_RDPIn + 
                     coeff_dict['KrOM_MiN_VT'] * Dt_rOMIn + 
                     coeff_dict['KForNDF_MiN_VT'] * Dt_ForNDFIn + 
                     coeff_dict['KrOM2_MiN_VT'] * Dt_rOMIn**2 + 
                     coeff_dict['KrdStxrOM_MiN_VT'] * Rum_DigStIn * Dt_rOMIn + 
                     coeff_dict['KrdNDFxForNDF_MiN_VT'] * Rum_DigNDFIn * 
                     Dt_ForNDFIn)
    return Du_MiN_VTln_g


def calculate_Du_MiN_VTnln_g(
    An_RDPIn: float, 
    Rum_DigNDFIn: float, 
    Rum_DigStIn: float
) -> float:
    Du_MiN_VTnln_g = (7.47 + 0.574 * An_RDPIn * 1000 / 
                      (1 + 3.60 / Rum_DigNDFIn + 12.3 / Rum_DigStIn)) # Line 1147
    return Du_MiN_VTnln_g


def calculate_Du_MiN_g(
    MiN_eqn: int, 
    MiN_Vm: float, 
    Dt_rOMIn: float,
    Dt_ForNDFIn: float,
    An_RDPIn: float,
    Rum_DigNDFIn: float,
    Rum_DigStIn: float,
    An_RDPIn_g: float,
    coeff_dict: dict
) -> float:
    if MiN_eqn == 1:
        Du_MiN_g = calculate_Du_MiN_NRC2021_g(
            MiN_Vm, Rum_DigNDFIn, Rum_DigStIn, An_RDPIn_g, 
            coeff_dict
            )
    elif MiN_eqn == 2:
        Du_MiN_g = calculate_Du_MiN_VTln_g(
            Dt_rOMIn, Dt_ForNDFIn, An_RDPIn, Rum_DigStIn, Rum_DigNDFIn, 
            coeff_dict
            )
    elif MiN_eqn == 3:
        Du_MiN_g = calculate_Du_MiN_VTnln_g(
            An_RDPIn, Rum_DigNDFIn, Rum_DigStIn
            )
    else:
        raise ValueError(
            f"Invalid MiN_eqn: {MiN_eqn} was entered. "
            "Must choose 1, 2 or 3."
            )
    return Du_MiN_g


def calculate_Du_MiCP(Du_MiCP_g: float) -> float:
    Du_MiCP = Du_MiCP_g / 1000  # Line 1166
    return Du_MiCP


def calculate_Du_idMiCP_g(Du_MiCP_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"SI_dcMiCP": 80}
    
    calculate_Du_idMiCP_g(
        Du_MiCP_g = 200.0, coeff_dict = coeff_dict
    )
    ```
    """
    Du_idMiCP_g = coeff_dict['SI_dcMiCP'] / 100 * Du_MiCP_g  # Line 1180
    return Du_idMiCP_g


def calculate_Du_idMiCP(Du_idMiCP_g: float) -> float:
    Du_idMiCP = Du_idMiCP_g / 1000
    return Du_idMiCP


def calculate_Du_idMiTP_g(Du_idMiCP_g: float, coeff_dict: dict) -> float:
    """
    Examples
    --------
    ```
    coeff_dict = {"fMiTP_MiCP": 0.824}
    
    calculate_Du_idMiTP_g(
        Du_idMiCP_g = 160.0, coeff_dict = coeff_dict
    )
    ```
    """
    Du_idMiTP_g = coeff_dict['fMiTP_MiCP'] * Du_idMiCP_g  # Line 1182
    return Du_idMiTP_g


def calculate_Du_idMiTP(Du_idMiTP_g: float) -> float:
    Du_idMiTP = Du_idMiTP_g / 1000
    return Du_idMiTP


def calculate_Du_MiTP(Du_MiTP_g: float) -> float:
    """
    Du_MiTP_g: Duodenal microbial true protein, g/d
    """
    Du_MiTP = Du_MiTP_g / 1000  # Line 1167
    return Du_MiTP


def calculate_Du_EndCP_g(Dt_DMIn: float, InfRum_DMIn: float) -> float:
    """
    Du_EndCP_g: Duodenal endogenous flow of crude protein, g/d
    """
    Du_EndCP_g = 96.1 + 7.54 * (Dt_DMIn + InfRum_DMIn)  # Line 1170
    return Du_EndCP_g


def calculate_Du_EndN_g(Dt_DMIn: float, InfRum_DMIn: float) -> float:
    """
    Du_EndN_g: Duodenal endogenous flow of nitrogen, g/d 
    """
    Du_EndN_g = 15.4 + 1.21 * (Dt_DMIn + InfRum_DMIn)
    # g/d, recalc of the above eqn (calculate_Du_EndCP_g) at 16% N in CP, Line 1171
    return Du_EndN_g


def calculate_Du_EndCP(Du_EndCP_g: float) -> float:
    """
    Du_EndCP: Duodenal endogenous flow of crude protein, kg/d 
    """
    Du_EndCP = Du_EndCP_g / 1000  # Line 1172
    return Du_EndCP


def calculate_Du_EndN(Du_EndN_g: float) -> float:
    """
    Du_EndN: Duodenal endogenous flow of nitrogen, kg/d 
    """
    Du_EndN = Du_EndN_g / 1000  # Line 1173
    return Du_EndN


def calculate_Du_NAN_g(
    Du_MiN_g: float, 
    An_RUPIn: float,
    Du_EndN_g: float
) -> float:
    """
    Du_NAN_g: ??? Not sure what NAN means. Something to do with Nitrogen flow. g/d
    """
    Du_NAN_g = Du_MiN_g + An_RUPIn * 0.16 * 1000 + Du_EndN_g  # Line 1175
    return Du_NAN_g


def calculate_Du_NANMN_g(An_RUPIn: float, Du_EndN_g: float) -> float:
    """
    Du_NANMN_g: ??? Not sure what NANMN means, Also has to do with Nitrogen flow, g/d
    """
    Du_NANMN_g = An_RUPIn * 0.16 * 1000 + Du_EndN_g  # Line 1176
    return Du_NANMN_g


def calculate_Du_MiN_NRC2001_g(Dt_TDNIn: float, An_RDPIn: float) -> float:
    """
    Du_MiN_NRC2001_g: Microbial N flow, NRC 2001 equation (g/d)
    """
    # NRC 2001 eqn. for N flow comparison purposes, Line 1389
    if 0.13 * Dt_TDNIn > 0.85 * An_RDPIn:
        Du_MiN_NRC2001_g = 0.85 * An_RDPIn * 1000 * 0.16
    else:
        Du_MiN_NRC2001_g = 0.13 * Dt_TDNIn * 1000 * 0.16
    return Du_MiN_NRC2001_g


def calculate_Rum_MiCP_DigCHO(
    Du_MiCP: float, 
    Rum_DigNDFIn: float,
    Rum_DigStIn: float
) -> float:
    """
    Rum_MiCP_DigCHO: Microbial CP as a fraction of digestable carbohydrates
    """
    Rum_MiCP_DigCHO = Du_MiCP / (Rum_DigNDFIn + Rum_DigStIn)  # Line 3122
    return Rum_MiCP_DigCHO
